# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import TYPE_CHECKING

from twisted.internet import threads
from twisted.python import threadpool

from buildbot.test.fake.reactor import NonThreadPool
from buildbot.test.fake.reactor import TestReactor
from buildbot.util import twisted
from buildbot.util.eventual import _setReactor
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from twisted.trial import unittest

    _TestReactorMixinBase = unittest.TestCase
else:
    _TestReactorMixinBase = object


class TestReactorMixin(_TestReactorMixinBase):
    """
    Mix this in to get TestReactor as self.reactor which is correctly cleaned up
    at the end
    """

    def setup_test_reactor(self, use_asyncio=False, auto_tear_down=True) -> None:
        if use_asyncio:
            warn_deprecated('4.2.0', 'use_asyncio=True is deprecated')
        if not auto_tear_down:
            warn_deprecated('4.2.0', 'auto_tear_down=False is deprecated')

        self.patch(threadpool, 'ThreadPool', NonThreadPool)
        self.patch(twisted, 'ThreadPool', NonThreadPool)
        self.reactor = TestReactor()
        self.reactor.set_test_case(self)

        _setReactor(self.reactor)

        def deferToThread(f, *args, **kwargs):
            return threads.deferToThreadPool(
                self.reactor, self.reactor.getThreadPool(), f, *args, **kwargs
            )

        self.patch(threads, 'deferToThread', deferToThread)

        if auto_tear_down:
            self.addCleanup(self.tear_down_test_reactor)
        self._reactor_tear_down_called = False

    def tear_down_test_reactor(self) -> None:
        if self._reactor_tear_down_called:
            return

        self._reactor_tear_down_called = True

        # During shutdown sequence we must first stop the reactor and only then set unset the
        # reactor used for eventually() because any callbacks that are run during reactor.stop()
        # may use eventually() themselves.
        self.reactor.stop()
        self.reactor.assert_no_remaining_calls()
        _setReactor(None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.config.checks import check_param_length
from buildbot.config.checks import check_param_str
from buildbot.config.checks import check_param_str_none
from buildbot.process.properties import Interpolate
from buildbot.test.util import config


class TestCheckParamLength(unittest.TestCase, config.ConfigErrorsMixin):
    def test_short_string(self):
        check_param_length('1234567890', 'Step name', 10)

    def test_long_string(self):
        with self.assertRaisesConfigError("exceeds maximum length of 10"):
            check_param_length('12345678901', 'Step name', 10)

    def test_short_interpolate(self):
        check_param_length(Interpolate('123456%(prop:xy)s7890'), 'Step name', 10)

    def test_short_interpolate_args(self):
        check_param_length(Interpolate('123456%s7890', 'arg'), 'Step name', 10)

    def test_short_interpolate_kwargs(self):
        check_param_length(Interpolate('123456%(prop:xy)s7890', kw='arg'), 'Step name', 10)

    def test_long_interpolate(self):
        with self.assertRaisesConfigError("exceeds maximum length of 10"):
            check_param_length(Interpolate('123456%(prop:xy)s78901'), 'Step name', 10)


class TestCheckParamType(unittest.TestCase, config.ConfigErrorsMixin):
    def test_str(self):
        check_param_str('abc', self.__class__, 'param')

    def test_str_wrong(self):
        msg = "TestCheckParamType argument param must be an instance of str"
        with self.assertRaisesConfigError(msg):
            check_param_str(1, self.__class__, 'param')

    def test_str_none(self):
        check_param_str_none('abc', self.__class__, 'param')
        check_param_str_none(None, self.__class__, 'param')

    def test_str_none_wrong(self):
        msg = "TestCheckParamType argument param must be an instance of str or None"
        with self.assertRaisesConfigError(msg):
            check_param_str_none(1, self.__class__, 'param')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import Any

from twisted.trial import unittest

from buildbot_worker.test.reactor import TestReactorMixin
from buildbot_worker.util import buffer_manager


class BufferManager(TestReactorMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.results_collected: list[list[tuple[str, Any]]] = []
        self.setup_test_reactor()

    def message_consumer(self, msg_data: list[tuple[str, Any]]) -> None:
        self.results_collected.append(msg_data)

    def assert_sent_messages(self, expected: list[list[tuple[str, Any]]]) -> None:
        self.assertEqual(self.results_collected, expected)
        self.results_collected = []

    def test_append_message_rc_fits_in_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("rc", 1)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("rc", 1)]])

    def test_append_message_log_in_one_msg(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        manager.append("log", ('log_test', ('text\n', [4], [0.0])))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([
            [
                ("stdout", ('12\n', [2], [1.0])),
                ("log", ('log_test', ('text\n', [4], [0.0]))),
            ]
        ])

    def test_append_message_rc_in_one_msg(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        manager.append("rc", 1)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("stdout", ('12\n', [2], [1.0])), ("rc", 1)]])

    def test_append_message_log_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        manager.append("log", ('log_test', ('tex\n', [4], [0.0])))
        self.assert_sent_messages([[("stdout", ('12\n', [2], [1.0]))]])

        manager.flush()
        self.assert_sent_messages([[("log", ('log_test', ('tex\n', [4], [0.0])))]])

    def test_append_message_rc_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        manager.append("rc", 1)
        self.assert_sent_messages([[("stdout", ('12\n', [2], [1.0]))]])

        manager.flush()
        self.assert_sent_messages([[("rc", 1)]])

    def test_append_two_messages_rc_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 15, 5)

        manager.append("rc", 1)
        self.assert_sent_messages([[("rc", 1)]])

        manager.append("rc", 0)
        self.assert_sent_messages([[("rc", 0)]])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_two_messages_rc_fits_in_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("rc", 1)
        self.assert_sent_messages([])

        manager.append("rc", 0)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("rc", 1), ("rc", 0)]])

    def test_append_message_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ('12345\n', [5], [1.0]))
        self.assert_sent_messages([[("stdout", ("12345\n", [5], [1.0]))]])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_message_fits_in_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 18, 5)

        manager.append("stdout", ("1\n", [1], [1.0]))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("stdout", ("1\n", [1], [1.0]))]])

        # only to see if flush does not send any more messages
        manager.flush()
        self.assert_sent_messages([])

    def test_append_two_messages_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ("1\n", [1], [1.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ("22\n", [2], [2.0]))
        self.assert_sent_messages([[("stdout", ("1\n", [1], [1.0]))]])

        manager.flush()
        self.assert_sent_messages([[('stdout', ('22\n', [2], [2.0]))]])

    def test_append_two_messages_same_logname_log_joined(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("log", ("log_test", ("1\n", [1], [1.0])))
        self.assert_sent_messages([])

        manager.append("log", ("log_test", ("2\n", [1], [2.0])))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("log", ("log_test", ("1\n2\n", [1, 3], [1.0, 2.0])))]])

    def test_append_two_messages_same_logname_joined(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("stdout", ("1\n", [1], [1.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ("2\n", [1], [2.0]))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([[("stdout", ("1\n2\n", [1, 3], [1.0, 2.0]))]])

    def test_append_two_messages_same_logname_log_joined_many_lines(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 80, 5)

        manager.append("log", ("log_test", ("1\n2\n", [1, 3], [1.0, 2.0])))
        self.assert_sent_messages([])

        manager.append("log", ("log_test", ("3\n4\n", [1, 3], [3.0, 4.0])))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([
            [("log", ("log_test", ("1\n2\n3\n4\n", [1, 3, 5, 7], [1.0, 2.0, 3.0, 4.0])))]
        ])

    def test_append_two_messages_same_logname_joined_many_lines(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 80, 5)

        manager.append("stdout", ("1\n2\n", [1, 3], [1.0, 2.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ("3\n4\n", [1, 3], [3.0, 4.0]))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([
            [("stdout", ("1\n2\n3\n4\n", [1, 3, 5, 7], [1.0, 2.0, 3.0, 4.0]))]
        ])

    def test_append_three_messages_not_same_logname_log_not_joined(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 70, 5)

        manager.append("log", ("log_test", ("1\n", [1], [1.0])))
        self.assert_sent_messages([])

        manager.append("log", ("log_test2", ("2\n", [1], [2.0])))
        self.assert_sent_messages([])

        manager.append("log", ("log_test3", ("3\n", [1], [3.0])))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([
            [
                ("log", ("log_test", ("1\n", [1], [1.0]))),
                ("log", ("log_test2", ("2\n", [1], [2.0]))),
                ("log", ("log_test3", ("3\n", [1], [3.0]))),
            ]
        ])

    def test_append_three_messages_not_same_logname_not_joined(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 60, 5)

        manager.append("stdout", ("1\n", [1], [1.0]))
        self.assert_sent_messages([])

        manager.append("stderr", ("2\n", [1], [2.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ("3\n", [1], [3.0]))
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([
            [
                ("stdout", ("1\n", [1], [1.0])),
                ("stderr", ("2\n", [1], [2.0])),
                ("stdout", ("3\n", [1], [3.0])),
            ]
        ])

    def test_append_two_messages_same_logname_log_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("log", ("log_test", ("1234\n", [4], [1.0])))
        self.assert_sent_messages([[("log", ("log_test", ("1234\n", [4], [1.0])))]])

        manager.append("log", ("log_test", ("5678\n", [4], [2.0])))
        self.assert_sent_messages([[("log", ("log_test", ("5678\n", [4], [2.0])))]])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_two_messages_same_logname_exceeds_buffer(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ("1234\n", [4], [1.0]))
        self.assert_sent_messages([[("stdout", ("1234\n", [4], [1.0]))]])

        manager.append("stdout", ("5678\n", [4], [2.0]))
        self.assert_sent_messages([[("stdout", ("5678\n", [4], [2.0]))]])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_exceeds_buffer_log_long_line_first_line_too_long(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append(
            "log",
            (
                "log_test",
                ("tbe5\nta4\ntbe5\nt3\ntd4\n", [4, 8, 13, 16, 20], [1.0, 2.0, 3.0, 4.0, 5.0]),
            ),
        )

        self.assert_sent_messages([
            [("log", ("log_test", ("tbe5\n", [4], [1.0])))],
            [("log", ("log_test", ("ta4\n", [3], [2.0])))],
            [("log", ("log_test", ("tbe5\n", [4], [3.0])))],
            [("log", ("log_test", ("t3\n", [2], [4.0])))],
            [("log", ("log_test", ("td4\n", [3], [5.0])))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_exceeds_buffer_long_line_first_line_too_long(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append(
            "stdout",
            ("tbe5\nta4\ntbe5\nt3\ntd4\n", [4, 8, 13, 16, 20], [1.0, 2.0, 3.0, 4.0, 5.0]),
        )

        self.assert_sent_messages([
            [("stdout", ("tbe5\n", [4], [1.0]))],
            [("stdout", ("ta4\n", [3], [2.0]))],
            [("stdout", ("tbe5\n", [4], [3.0]))],
            [("stdout", ("t3\n", [2], [4.0]))],
            [("stdout", ("td4\n", [3], [5.0]))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_exceeds_buffer_log_long_line_middle_line_too_long(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append(
            "log",
            (
                "log_test",
                ("t3\nta4\ntbe5\nt3\ntd4\n", [2, 6, 11, 14, 18], [1.0, 2.0, 3.0, 4.0, 5.0]),
            ),
        )

        self.assert_sent_messages([
            [("log", ("log_test", ("t3\n", [2], [1.0])))],
            [("log", ("log_test", ("ta4\n", [3], [2.0])))],
            [("log", ("log_test", ("tbe5\n", [4], [3.0])))],
            [("log", ("log_test", ("t3\n", [2], [4.0])))],
            [("log", ("log_test", ("td4\n", [3], [5.0])))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_exceeds_buffer_long_line_middle_line_too_long(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append(
            "stdout",
            ("t3\nta4\ntbe5\nt3\ntd4\n", [2, 6, 11, 14, 18], [1.0, 2.0, 3.0, 4.0, 5.0]),
        )

        self.assert_sent_messages([
            [("stdout", ("t3\n", [2], [1.0]))],
            [("stdout", ("ta4\n", [3], [2.0]))],
            [("stdout", ("tbe5\n", [4], [3.0]))],
            [("stdout", ("t3\n", [2], [4.0]))],
            [("stdout", ("td4\n", [3], [5.0]))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_long_line_log_concatenate(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 60, 5)

        manager.append(
            "log",
            (
                "log_test",
                (
                    "text_text_text_text_text_text_text_text_\nlex\nteym\nte\ntuz\n",
                    [40, 44, 49, 52, 56],
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                ),
            ),
        )

        self.assert_sent_messages([
            [("log", ("log_test", ("text_text_text_text_text_text_text_text_\n", [40], [1.0])))],
            [("log", ("log_test", ("lex\nteym\nte\n", [3, 8, 11], [2.0, 3.0, 4.0])))],
            [("log", ("log_test", ("tuz\n", [3], [5.0])))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_long_line_concatenate(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 60, 5)

        manager.append(
            "stdout",
            (
                "text_text_text_text_text_text_text_text_\nlex\nteym\nte\ntuz\n",
                [40, 44, 49, 52, 56],
                [1.0, 2.0, 3.0, 4.0, 5.0],
            ),
        )

        self.assert_sent_messages([
            [("stdout", ("text_text_text_text_text_text_text_text_\n", [40], [1.0]))],
            [("stdout", ("lex\nteym\nte\n", [3, 8, 11], [2.0, 3.0, 4.0]))],
            [("stdout", ("tuz\n", [3], [5.0]))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_log_not_fitting_line_after_fitting_line(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("log", ("log_test", ("12\n", [4], [1.0])))
        self.assert_sent_messages([])

        manager.append("log", ("log_test", ("345678\n", [6], [2.0])))
        self.assert_sent_messages([
            [("log", ("log_test", ("12\n", [4], [1.0])))],
            [("log", ("log_test", ("345678\n", [6], [2.0])))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_not_fitting_line_after_fitting_line(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ("12\n", [4], [1.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ("345678\n", [6], [2.0]))
        self.assert_sent_messages([
            [("stdout", ("12\n", [4], [1.0]))],
            [("stdout", ("345678\n", [6], [2.0]))],
        ])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_timeout_fits_in_buffer_timeout_expires_with_message(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        self.reactor.advance(4)
        self.assert_sent_messages([])

        self.reactor.advance(1)
        self.assert_sent_messages([[("stdout", ("12\n", [2], [1.0]))]])

        self.reactor.advance(5)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_timeout_fits_in_buffer_two_messages_before_timeout_expires(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        self.reactor.advance(1)
        manager.append("stdout", ('345\n', [3], [2.0]))
        self.assert_sent_messages([])

        self.reactor.advance(4)
        self.assert_sent_messages([[("stdout", ("12\n345\n", [2, 6], [1.0, 2.0]))]])

        self.reactor.advance(5)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_timeout_two_messages_timeout_expires_with_single_message(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 40, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        self.reactor.advance(5)
        self.assert_sent_messages([[("stdout", ("12\n", [2], [1.0]))]])

        manager.append("stdout", ('345\n', [3], [2.0]))
        self.assert_sent_messages([])

        self.reactor.advance(5)
        self.assert_sent_messages([[("stdout", ("345\n", [3], [2.0]))]])

        manager.flush()
        self.assert_sent_messages([])

    def test_append_timeout_long_line_flushes_short_line_before_timeout(self) -> None:
        manager = buffer_manager.BufferManager(self.reactor, self.message_consumer, 20, 5)

        manager.append("stdout", ('12\n', [2], [1.0]))
        self.assert_sent_messages([])

        manager.append("stdout", ('345678\n', [6], [2.0]))
        self.assert_sent_messages([
            [("stdout", ("12\n", [2], [1.0]))],
            [("stdout", ("345678\n", [6], [2.0]))],
        ])

        self.reactor.advance(5)
        self.assert_sent_messages([])

        manager.flush()
        self.assert_sent_messages([])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import warnings
from unittest.mock import Mock
from unittest.mock import call

from packaging.version import parse as parse_version
from parameterized import parameterized
from twisted.internet import defer
from twisted.internet import error
from twisted.internet import reactor
from twisted.python import failure
from twisted.trial import unittest

from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.reporters import utils
from buildbot.reporters.generators.build import BuildStartEndStatusGenerator
from buildbot.reporters.generators.buildset import BuildSetStatusGenerator
from buildbot.reporters.gerrit import GERRIT_LABEL_REVIEWED
from buildbot.reporters.gerrit import GERRIT_LABEL_VERIFIED
from buildbot.reporters.gerrit import GerritBuildEndStatusGenerator
from buildbot.reporters.gerrit import GerritBuildSetStatusGenerator
from buildbot.reporters.gerrit import GerritBuildStartStatusGenerator
from buildbot.reporters.gerrit import GerritStatusPush
from buildbot.reporters.gerrit import defaultReviewCB
from buildbot.reporters.gerrit import defaultSummaryCB
from buildbot.reporters.gerrit import extract_project_revision
from buildbot.reporters.gerrit import makeReviewResult
from buildbot.reporters.message import MessageFormatterFunctionRaw
from buildbot.reporters.message import MessageFormatterRenderable
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.reporter import ReporterTestMixin

warnings.filterwarnings('error', message='.*Gerrit status')


def sampleReviewCB(builderName, build, result, status, arg):
    verified = 1 if result == SUCCESS else -1
    return makeReviewResult(
        str({'name': builderName, 'result': result}), (GERRIT_LABEL_VERIFIED, verified)
    )


@defer.inlineCallbacks
def sampleReviewCBDeferred(builderName, build, result, status, arg):
    verified = 1 if result == SUCCESS else -1
    result = yield makeReviewResult(
        str({'name': builderName, 'result': result}), (GERRIT_LABEL_VERIFIED, verified)
    )
    return result


def sampleStartCB(builderName, build, arg):
    return makeReviewResult(str({'name': builderName}), (GERRIT_LABEL_REVIEWED, 0))


@defer.inlineCallbacks
def sampleStartCBDeferred(builderName, build, arg):
    result = yield makeReviewResult(str({'name': builderName}), (GERRIT_LABEL_REVIEWED, 0))
    return result


def sampleSummaryCB(buildInfoList, results, status, arg):
    success = False
    failure = False

    for buildInfo in buildInfoList:
        if buildInfo['result'] == SUCCESS:  # pylint: disable=simplifiable-if-statement
            success = True
        else:
            failure = True

    if failure:
        verified = -1
    elif success:
        verified = 1
    else:
        verified = 0

    return makeReviewResult(str(buildInfoList), (GERRIT_LABEL_VERIFIED, verified))


@defer.inlineCallbacks
def sampleSummaryCBDeferred(buildInfoList, results, master, arg):
    success = False
    failure = False

    for buildInfo in buildInfoList:
        if buildInfo['result'] == SUCCESS:  # pylint: disable=simplifiable-if-statement
            success = True
        else:
            failure = True

    if failure:
        verified = -1
    elif success:
        verified = 1
    else:
        verified = 0

    result = yield makeReviewResult(str(buildInfoList), (GERRIT_LABEL_VERIFIED, verified))
    return result


class TestGerritStatusPush(TestReactorMixin, unittest.TestCase, ReporterTestMixin):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def setupGerritStatusPushSimple(self, *args, **kwargs):
        serv = kwargs.pop("server", "serv")
        username = kwargs.pop("username", "user")
        gsp = GerritStatusPush(serv, username, *args, **kwargs)
        yield gsp.setServiceParent(self.master)
        yield gsp.startService()
        return gsp

    @defer.inlineCallbacks
    def setupGerritStatusPush(self, *args, **kwargs):
        gsp = yield self.setupGerritStatusPushSimple(*args, **kwargs)
        gsp.send_code_review = Mock()
        return gsp

    @defer.inlineCallbacks
    def setupBuildResults(self, buildResults, finalResult):
        yield self.insert_test_data(buildResults, finalResult)
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True)
        builds = res['builds']
        buildset = res['buildset']

        @defer.inlineCallbacks
        def getChangesForBuild(buildid):
            assert buildid == 20
            ch = yield self.master.db.changes.getChange(13)
            return [ch]

        self.master.db.changes.getChangesForBuild = getChangesForBuild
        return (buildset, builds)

    def makeBuildInfo(self, buildResults, resultText, builds):
        info = []
        for i, buildResult in enumerate(buildResults):
            info.append({
                'name': f"Builder{i}",
                'result': buildResult,
                'resultText': resultText[i],
                'text': 'buildText',
                'url': f"http://localhost:8080/#/builders/{79 + i}/builds/{i}",
                'build': builds[i],
            })
        return info

    @defer.inlineCallbacks
    def run_fake_summary_build(self, gsp, buildResults, finalResult, resultText, expWarning=False):
        buildset, builds = yield self.setupBuildResults(buildResults, finalResult)
        yield gsp._got_event(("buildsets", 98, "complete"), buildset)

        info = self.makeBuildInfo(buildResults, resultText, builds)
        if expWarning:
            self.assertEqual(
                [w['message'] for w in self.flushWarnings()],
                [
                    'The Gerrit status callback uses the old '
                    'way to communicate results.  The outcome '
                    'might be not what is expected.'
                ],
            )
        return str(info)

    @defer.inlineCallbacks
    def check_summary_build_deferred(self, buildResults, finalResult, resultText, verifiedScore):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                GerritBuildSetStatusGenerator(
                    callback=sampleSummaryCBDeferred,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                )
            ]
        )

        msg = yield self.run_fake_summary_build(gsp, buildResults, finalResult, resultText)

        gsp.send_code_review.assert_called_once_with(
            self.reporter_test_project,
            self.reporter_test_revision,
            msg,
            {GERRIT_LABEL_VERIFIED: verifiedScore},
        )

    @defer.inlineCallbacks
    def check_summary_build(self, buildResults, finalResult, resultText, verifiedScore):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                GerritBuildSetStatusGenerator(
                    callback=sampleSummaryCB,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                )
            ]
        )

        msg = yield self.run_fake_summary_build(gsp, buildResults, finalResult, resultText)

        gsp.send_code_review.assert_called_once_with(
            self.reporter_test_project,
            self.reporter_test_revision,
            msg,
            {GERRIT_LABEL_VERIFIED: verifiedScore},
        )

    @defer.inlineCallbacks
    def test_gerrit_ssh_cmd(self):
        kwargs = {
            'server': 'example.com',
            'username': 'buildbot',
        }
        without_identity = yield self.setupGerritStatusPush(**kwargs)

        expected1 = [
            'ssh',
            '-o',
            'BatchMode=yes',
            'buildbot@example.com',
            '-p',
            '29418',
            'gerrit',
            'foo',
        ]
        self.assertEqual(expected1, without_identity._gerritCmd('foo'))
        yield without_identity.disownServiceParent()
        with_identity = yield self.setupGerritStatusPush(identity_file='/path/to/id_rsa', **kwargs)
        expected2 = [
            'ssh',
            '-o',
            'BatchMode=yes',
            '-i',
            '/path/to/id_rsa',
            'buildbot@example.com',
            '-p',
            '29418',
            'gerrit',
            'foo',
        ]
        self.assertEqual(expected2, with_identity._gerritCmd('foo'))

    def test_buildsetComplete_success_sends_summary_review_deferred(self):
        d = self.check_summary_build_deferred(
            buildResults=[SUCCESS, SUCCESS],
            finalResult=SUCCESS,
            resultText=["succeeded", "succeeded"],
            verifiedScore=1,
        )
        return d

    def test_buildsetComplete_success_sends_summary_review(self):
        d = self.check_summary_build(
            buildResults=[SUCCESS, SUCCESS],
            finalResult=SUCCESS,
            resultText=["succeeded", "succeeded"],
            verifiedScore=1,
        )
        return d

    def test_buildsetComplete_failure_sends_summary_review(self):
        d = self.check_summary_build(
            buildResults=[FAILURE, FAILURE],
            finalResult=FAILURE,
            resultText=["failed", "failed"],
            verifiedScore=-1,
        )
        return d

    def test_buildsetComplete_mixed_sends_summary_review(self):
        d = self.check_summary_build(
            buildResults=[SUCCESS, FAILURE],
            finalResult=FAILURE,
            resultText=["succeeded", "failed"],
            verifiedScore=-1,
        )
        return d

    @parameterized.expand([
        ("matched", ["Builder1"], True),
        ("not_matched", ["foo"], False),
    ])
    @defer.inlineCallbacks
    def test_buildset_complete_filtered_builder(self, name, builders, should_call):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                GerritBuildSetStatusGenerator(
                    callback=sampleSummaryCB,
                    callback_arg=None,
                    builders=builders,
                    want_steps=False,
                    want_logs=False,
                )
            ]
        )
        yield self.run_fake_summary_build(gsp, [FAILURE, FAILURE], FAILURE, ["failed", "failed"])

        self.assertEqual(gsp.send_code_review.called, should_call)

    @defer.inlineCallbacks
    def run_fake_single_build(self, gsp, buildResult, expWarning=False):
        _, builds = yield self.setupBuildResults([None], None)

        yield gsp._got_event(('builds', builds[0]['buildid'], 'new'), builds[0])

        yield self.master.db.builds.finishBuild(builds[0]["buildid"], buildResult)
        yield self.master.db.buildsets.completeBuildset(98, buildResult)

        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True)
        builds = res['builds']

        yield gsp._got_event(('builds', builds[0]['buildid'], 'finished'), builds[0])

        if expWarning:
            self.assertEqual(
                [w['message'] for w in self.flushWarnings()],
                [
                    'The Gerrit status callback uses the old '
                    'way to communicate results.  The outcome '
                    'might be not what is expected.'
                ],
            )

        return str({'name': 'Builder0', 'result': buildResult})

    # same goes for check_single_build
    @defer.inlineCallbacks
    def check_single_build(self, buildResult, verifiedScore):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                GerritBuildEndStatusGenerator(
                    callback=sampleReviewCB,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                ),
                GerritBuildStartStatusGenerator(
                    callback=sampleStartCB,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                ),
            ]
        )

        msg = yield self.run_fake_single_build(gsp, buildResult)
        calls = [
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                str({'name': self.reporter_test_builder_name}),
                {GERRIT_LABEL_REVIEWED: 0},
            ),
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                msg,
                {GERRIT_LABEL_VERIFIED: verifiedScore},
            ),
        ]
        gsp.send_code_review.assert_has_calls(calls)

    def test_buildComplete_success_sends_review(self):
        return self.check_single_build(SUCCESS, 1)

    def test_buildComplete_failure_sends_review(self):
        return self.check_single_build(FAILURE, -1)

    # same goes for check_single_build
    @parameterized.expand([
        ("matched", ["Builder0"], True),
        ("not_matched", ["foo"], False),
    ])
    @defer.inlineCallbacks
    def test_single_build_filtered(self, name, builders, should_call):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                GerritBuildEndStatusGenerator(
                    callback=sampleReviewCB,
                    callback_arg=None,
                    builders=builders,
                    want_steps=False,
                    want_logs=False,
                ),
                GerritBuildStartStatusGenerator(
                    callback=sampleStartCB,
                    callback_arg=None,
                    builders=builders,
                    want_steps=False,
                    want_logs=False,
                ),
            ]
        )

        yield self.run_fake_single_build(gsp, SUCCESS)
        self.assertEqual(gsp.send_code_review.called, should_call)

    @parameterized.expand([
        ("success", SUCCESS, 1),
        ("failure", FAILURE, -1),
    ])
    @defer.inlineCallbacks
    def test_single_build_generators(self, name, build_result, verified_score):
        gsp = yield self.setupGerritStatusPush(generators=[BuildStartEndStatusGenerator()])

        yield self.run_fake_single_build(gsp, build_result)
        calls = [
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                "Build started.",
                {GERRIT_LABEL_VERIFIED: 0},
            ),
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                "Build done.",
                {GERRIT_LABEL_VERIFIED: verified_score},
            ),
        ]
        gsp.send_code_review.assert_has_calls(calls)

    @parameterized.expand([
        ("success", SUCCESS, 1),
        ("failure", FAILURE, -1),
    ])
    @defer.inlineCallbacks
    def test_single_buildset_generators(self, name, build_result, verified_score):
        gsp = yield self.setupGerritStatusPush(
            generators=[
                BuildSetStatusGenerator(message_formatter=MessageFormatterRenderable("Build done."))
            ]
        )

        yield self.run_fake_summary_build(gsp, [build_result], build_result, "text")
        calls = [
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                "Build done.",
                {GERRIT_LABEL_VERIFIED: verified_score},
            )
        ]
        gsp.send_code_review.assert_has_calls(calls)

    @defer.inlineCallbacks
    def test_single_buildset_generators_override_label(self):
        formatter = MessageFormatterFunctionRaw(
            lambda _, __: {
                "body": "text1",
                "type": "plain",
                "subject": "sub1",
                "extra_info": {"labels": {"Verified": -2}},
            }
        )

        gsp = yield self.setupGerritStatusPush(
            generators=[BuildSetStatusGenerator(message_formatter=formatter)]
        )

        yield self.run_fake_summary_build(gsp, [SUCCESS], SUCCESS, "text")
        calls = [
            call(
                self.reporter_test_project,
                self.reporter_test_revision,
                "text1",
                {GERRIT_LABEL_VERIFIED: -2},
            )
        ]
        gsp.send_code_review.assert_has_calls(calls)

    def test_defaultReviewCBSuccess(self):
        res = defaultReviewCB("builderName", {}, SUCCESS, None, None)
        self.assertEqual(res['labels'], {'Verified': 1})
        res = defaultReviewCB("builderName", {}, RETRY, None, None)
        self.assertEqual(res['labels'], {})

    def test_defaultSummaryCB(self):
        info = self.makeBuildInfo([SUCCESS, FAILURE], ["yes", "no"], [None, None])
        res = defaultSummaryCB(info, SUCCESS, None, None)
        self.assertEqual(res['labels'], {'Verified': -1})
        info = self.makeBuildInfo([SUCCESS, SUCCESS], ["yes", "yes"], [None, None])
        res = defaultSummaryCB(info, SUCCESS, None, None)
        self.assertEqual(res['labels'], {'Verified': 1})

    @defer.inlineCallbacks
    def testBuildGerritCommand(self):
        gsp = yield self.setupGerritStatusPushSimple()
        spawnSkipFirstArg = Mock()
        gsp.spawnProcess = lambda _, *a, **k: spawnSkipFirstArg(*a, **k)
        yield gsp.send_code_review("project", "revision", "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            ['ssh', '-o', 'BatchMode=yes', 'user@serv', '-p', '29418', 'gerrit', 'version'],
            env=None,
        )
        gsp.processVersion(parse_version("2.6"), lambda: None)
        spawnSkipFirstArg = Mock()
        yield gsp.send_code_review("project", "revision", "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            [
                'ssh',
                '-o',
                'BatchMode=yes',
                'user@serv',
                '-p',
                '29418',
                'gerrit',
                'review',
                '--project project',
                "--message 'bla'",
                '--label Verified=1',
                'revision',
            ],
            env=None,
        )

        # <=2.5 uses other syntax
        gsp.processVersion(parse_version("2.4"), lambda: None)
        spawnSkipFirstArg = Mock()
        yield gsp.send_code_review("project", "revision", "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            [
                'ssh',
                '-o',
                'BatchMode=yes',
                'user@serv',
                '-p',
                '29418',
                'gerrit',
                'review',
                '--project project',
                "--message 'bla'",
                '--verified 1',
                'revision',
            ],
            env=None,
        )

        # now test the notify argument, even though _gerrit_notify
        # is private, work around that
        gsp._gerrit_notify = 'OWNER'
        gsp.processVersion(parse_version('2.6'), lambda: None)
        spawnSkipFirstArg = Mock()
        yield gsp.send_code_review('project', 'revision', "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            [
                'ssh',
                '-o',
                'BatchMode=yes',
                'user@serv',
                '-p',
                '29418',
                'gerrit',
                'review',
                '--project project',
                '--notify OWNER',
                "--message 'bla'",
                '--label Verified=1',
                'revision',
            ],
            env=None,
        )

        # gerrit versions <= 2.5 uses other syntax
        gsp.processVersion(parse_version('2.4'), lambda: None)
        spawnSkipFirstArg = Mock()
        yield gsp.send_code_review('project', 'revision', "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            [
                'ssh',
                '-o',
                'BatchMode=yes',
                'user@serv',
                '-p',
                '29418',
                'gerrit',
                'review',
                '--project project',
                '--notify OWNER',
                "--message 'bla'",
                '--verified 1',
                'revision',
            ],
            env=None,
        )

        gsp.processVersion(parse_version("2.13"), lambda: None)
        spawnSkipFirstArg = Mock()
        yield gsp.send_code_review("project", "revision", "bla", {'Verified': 1})
        spawnSkipFirstArg.assert_called_once_with(
            'ssh',
            [
                'ssh',
                '-o',
                'BatchMode=yes',
                'user@serv',
                '-p',
                '29418',
                'gerrit',
                'review',
                '--project project',
                '--tag autogenerated:buildbot',
                '--notify OWNER',
                "--message 'bla'",
                '--label Verified=1',
                'revision',
            ],
            env=None,
        )

    @defer.inlineCallbacks
    def test_callWithVersion_bytes_output(self):
        gsp = yield self.setupGerritStatusPushSimple()
        exp_argv = ['ssh', '-o', 'BatchMode=yes', 'user@serv', '-p', '29418', 'gerrit', 'version']

        def spawnProcess(pp, cmd, argv, env):
            self.assertEqual([cmd, argv], [exp_argv[0], exp_argv])
            pp.errReceived(b'test stderr\n')
            pp.outReceived(b'gerrit version 2.14\n')
            pp.outReceived(b'(garbage that should not cause a crash)\n')
            so = error.ProcessDone(None)
            pp.processEnded(failure.Failure(so))

        self.patch(reactor, 'spawnProcess', spawnProcess)
        gsp.callWithVersion(lambda: self.assertEqual(gsp.gerrit_version, parse_version('2.14')))

    def test_name_as_class_attribute(self):
        class FooStatusPush(GerritStatusPush):
            name = 'foo'

        reporter = FooStatusPush('gerrit.server.com', 'password')
        self.assertEqual(reporter.name, 'foo')

    def test_name_as_kwarg(self):
        reporter = GerritStatusPush('gerrit.server.com', 'password', name='foo')
        self.assertEqual(reporter.name, 'foo')

    def test_default_name(self):
        reporter = GerritStatusPush('gerrit.server.com', 'password')
        self.assertEqual(reporter.name, 'GerritStatusPush')

    @defer.inlineCallbacks
    def test_extract_project_revision(self):
        yield self.insert_test_data([SUCCESS], SUCCESS)
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True)
        report = {"builds": res["builds"], "buildset": res["buildset"]}

        project, revision = yield extract_project_revision(self.master, report)
        self.assertEqual(project, "testProject")
        self.assertEqual(revision, "d34db33fd43db33f")

    @defer.inlineCallbacks
    def test_extract_project_revision_no_build(self):
        yield self.insert_test_data([], SUCCESS)
        yield self.master.db.insert_test_data([
            fakedb.BuildsetProperty(
                buildsetid=98, property_name="event.change.id", property_value='["12345", "fakedb"]'
            ),
            fakedb.BuildsetProperty(
                buildsetid=98,
                property_name="event.change.project",
                property_value='["project1", "fakedb"]',
            ),
            fakedb.BuildsetProperty(
                buildsetid=98,
                property_name="event.patchSet.revision",
                property_value='["abcdabcd", "fakedb"]',
            ),
        ])
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True)
        report = {"builds": res["builds"], "buildset": res["buildset"]}

        project, revision = yield extract_project_revision(self.master, report)
        self.assertEqual(project, "project1")
        self.assertEqual(revision, "abcdabcd")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import random
from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.python import failure
from twisted.trial import unittest

from buildbot import config
from buildbot.db import buildrequests
from buildbot.process import buildrequestdistributor
from buildbot.process import factory
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import epoch2datetime
from buildbot.util.eventual import fireEventually
from buildbot.util.twisted import async_to_deferred


def nth_worker(n):
    def pick_nth_by_name(builder, workers=None, br=None):
        if workers is None:
            workers = builder
        workers = workers[:]
        workers.sort(key=lambda a: a.name)
        return workers[n]

    return pick_nth_by_name


class TestBRDBase(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.botmaster = mock.Mock(name='botmaster')
        self.botmaster.builders = {}
        self.builders = {}

        def prioritizeBuilders(master, builders):
            # simple sort-by-name by default
            return sorted(builders, key=lambda b1: b1.name)

        self.master = self.botmaster.master = yield fakemaster.make_master(
            self, wantData=True, wantDb=True
        )
        self.master.caches = fakemaster.FakeCaches()
        self.master.config.prioritizeBuilders = prioritizeBuilders
        self.brd = buildrequestdistributor.BuildRequestDistributor(self.botmaster)
        self.brd.parent = self.botmaster
        self.brd.startService()

        @defer.inlineCallbacks
        def cleanup():
            if self.brd.running:
                yield self.brd.stopService()

        self.addCleanup(cleanup)

        # a collection of rows that would otherwise clutter up every test
        self.base_rows = [
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            fakedb.SourceStamp(id=21),
            fakedb.Builder(id=77, name='A'),
            fakedb.Buildset(id=11, reason='because'),
            fakedb.BuildsetSourceStamp(sourcestampid=21, buildsetid=11),
        ]

    def make_workers(self, worker_count):
        rows = self.base_rows[:]
        for i in range(worker_count):
            self.addWorkers({f'test-worker{i}': 1})
            rows.append(fakedb.Buildset(id=100 + i, reason='because'))
            rows.append(fakedb.BuildsetSourceStamp(buildsetid=100 + i, sourcestampid=21))
            rows.append(fakedb.BuildRequest(id=10 + i, buildsetid=100 + i, builderid=77))
        return rows

    def addWorkers(self, workerforbuilders):
        """C{workerforbuilders} maps name : available"""
        for name, avail in workerforbuilders.items():
            wfb = mock.Mock(spec=['isAvailable'], name=name)
            wfb.name = name
            wfb.isAvailable.return_value = avail
            for bldr in self.builders.values():
                bldr.workers.append(wfb)

    @defer.inlineCallbacks
    def createBuilder(self, name, builderid=None, builder_config=None):
        if builderid is None:
            b = fakedb.Builder(name=name)
            yield self.master.db.insert_test_data([b])
            builderid = b.id

        bldr = mock.Mock(name=name)
        bldr.name = name
        self.botmaster.builders[name] = bldr
        self.builders[name] = bldr

        def maybeStartBuild(worker, builds):
            worker.isAvailable.return_value = False
            self.startedBuilds.append((worker.name, builds))
            d = defer.Deferred()
            self.reactor.callLater(0, d.callback, True)
            return d

        bldr.maybeStartBuild = maybeStartBuild
        bldr.getCollapseRequestsFn = lambda: False

        bldr.workers = []
        bldr.getAvailableWorkers = lambda: [w for w in bldr.workers if w.isAvailable()]
        bldr.getBuilderId = lambda: (builderid)
        if builder_config is None:
            bldr.config.nextWorker = None
            bldr.config.nextBuild = None
        else:
            bldr.config = builder_config

        def canStartBuild(*args):
            can = bldr.config.canStartBuild
            return not can or can(*args)

        bldr.canStartBuild = canStartBuild

        return bldr

    @defer.inlineCallbacks
    def addBuilders(self, names):
        self.startedBuilds = []

        for name in names:
            yield self.createBuilder(name)

    @defer.inlineCallbacks
    def assert_claims(self, brids):
        brs = yield self.master.data.get(('buildrequests',))
        got_brids = [
            br['buildrequestid']
            for br in brs
            if br['claimed_by_masterid'] == fakedb.FakeDBConnector.MASTER_ID
        ]

        self.assertEqual(list(set(got_brids)), list(set(brids)))


class Test(TestBRDBase):
    def checkAllCleanedUp(self):
        # check that the BRD didn't end with a stuck lock or in the 'active' state (which would mean
        # it ended without unwinding correctly)
        self.assertEqual(self.brd.pending_builders_lock.locked, False)
        self.assertEqual(self.brd.activity_lock.locked, False)
        self.assertEqual(self.brd.active, False)

    def useMock_maybeStartBuildsOnBuilder(self):
        # sets up a mock "maybeStartBuildsOnBuilder" so we can track
        # how the method gets invoked

        # keep track of the calls to brd.maybeStartBuildsOnBuilder
        self.maybeStartBuildsOnBuilder_calls = []

        def maybeStartBuildsOnBuilder(bldr):
            self.assertIdentical(self.builders[bldr.name], bldr)
            self.maybeStartBuildsOnBuilder_calls.append(bldr.name)
            return fireEventually()

        self.brd._maybeStartBuildsOnBuilder = maybeStartBuildsOnBuilder

    def removeBuilder(self, name):
        del self.builders[name]
        del self.botmaster.builders[name]

    # tests

    @defer.inlineCallbacks
    def test_maybeStartBuildsOn_simple(self):
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(['bldr1'])
        yield self.brd.maybeStartBuildsOn(['bldr1'])

        yield self.brd._waitForFinish()
        self.assertEqual(self.maybeStartBuildsOnBuilder_calls, ['bldr1'])
        self.checkAllCleanedUp()

    @defer.inlineCallbacks
    def test_maybeStartBuildsOn_parallel(self):
        # test 15 "parallel" invocations of maybeStartBuildsOn, with a
        # _sortBuilders that takes a while.  This is a regression test for bug
        # 1979.
        builders = [f'bldr{i:02}' for i in range(15)]

        def slow_sorter(master, bldrs):
            bldrs.sort(key=lambda b1: b1.name)
            d = defer.Deferred()
            self.reactor.callLater(0, d.callback, bldrs)

            def done(_):
                return _

            d.addCallback(done)
            return d

        self.master.config.prioritizeBuilders = slow_sorter

        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(builders)
        for bldr in builders:
            yield self.brd.maybeStartBuildsOn([bldr])

        yield self.brd._waitForFinish()
        self.assertEqual(self.maybeStartBuildsOnBuilder_calls, builders)
        self.checkAllCleanedUp()

    @defer.inlineCallbacks
    def test_maybeStartBuildsOn_exception(self):
        self.addBuilders(['bldr1'])

        async def _maybeStartBuildsOnBuilder(n):
            # fail slowly, so that the activity loop doesn't exit too soon
            d = defer.Deferred()
            self.reactor.callLater(0, d.errback, failure.Failure(RuntimeError("oh noes")))
            await d

        self.brd._maybeStartBuildsOnBuilder = _maybeStartBuildsOnBuilder

        yield self.brd.maybeStartBuildsOn(['bldr1'])

        yield self.brd._waitForFinish()
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)
        self.checkAllCleanedUp()

    @defer.inlineCallbacks
    def test_maybeStartBuildsOn_collapsing(self):
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(['bldr1', 'bldr2', 'bldr3'])
        yield self.brd.maybeStartBuildsOn(['bldr3'])
        yield self.brd.maybeStartBuildsOn(['bldr2', 'bldr1'])
        yield self.brd.maybeStartBuildsOn(['bldr4'])  # should be ignored
        yield self.brd.maybeStartBuildsOn(['bldr2'])  # already queued - ignored
        yield self.brd.maybeStartBuildsOn(['bldr3', 'bldr2'])

        yield self.brd._waitForFinish()
        # bldr3 gets invoked twice, since it's considered to have started
        # already when the first call to maybeStartBuildsOn returns
        self.assertEqual(self.maybeStartBuildsOnBuilder_calls, ['bldr3', 'bldr1', 'bldr2', 'bldr3'])
        self.checkAllCleanedUp()

    @defer.inlineCallbacks
    def test_maybeStartBuildsOn_builders_missing(self):
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(['bldr1', 'bldr2', 'bldr3'])
        yield self.brd.maybeStartBuildsOn(['bldr1', 'bldr2', 'bldr3'])
        # bldr1 is already run, so surreptitiously remove the other
        # two - nothing should crash, but the builders should not run
        self.removeBuilder('bldr2')
        self.removeBuilder('bldr3')

        yield self.brd._waitForFinish()
        self.assertEqual(self.maybeStartBuildsOnBuilder_calls, ['bldr1'])
        self.checkAllCleanedUp()

    @defer.inlineCallbacks
    def do_test_sortBuilders(
        self,
        prioritizeBuilders,
        oldestRequestTimes,
        highestPriorities,
        expected,
        returnDeferred=False,
    ):
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(list(oldestRequestTimes))
        self.master.config.prioritizeBuilders = prioritizeBuilders

        def mklambda(t):  # work around variable-binding issues
            if returnDeferred:
                return lambda: defer.succeed(t)
            return lambda: t

        for n, t in oldestRequestTimes.items():
            if t is not None:
                t = epoch2datetime(t)
            self.builders[n].getOldestRequestTime = mklambda(t)

        for n, t in highestPriorities.items():
            self.builders[n].get_highest_priority = mklambda(t)

        result = yield self.brd._sortBuilders(list(oldestRequestTimes))

        self.assertEqual(result, expected)
        self.checkAllCleanedUp()

    def test_sortBuilders_default_sync(self):
        return self.do_test_sortBuilders(
            None,  # use the default sort
            {"bldr1": 777, "bldr2": 999, "bldr3": 888},
            {"bldr1": 10, "bldr2": 15, "bldr3": 5},
            ['bldr2', 'bldr1', 'bldr3'],
        )

    def test_sortBuilders_default_asyn(self):
        return self.do_test_sortBuilders(
            None,  # use the default sort
            {"bldr1": 777, "bldr2": 999, "bldr3": 888},
            {"bldr1": 10, "bldr2": 15, "bldr3": 5},
            ['bldr2', 'bldr1', 'bldr3'],
            returnDeferred=True,
        )

    def test_sortBuilders_default_None(self):
        return self.do_test_sortBuilders(
            None,  # use the default sort
            {"bldr1": 777, "bldr2": None, "bldr3": 888},
            {"bldr1": 10, "bldr2": None, "bldr3": 5},
            ['bldr1', 'bldr3', 'bldr2'],
        )

    def test_sortBuilders_default_priority_match(self):
        return self.do_test_sortBuilders(
            None,  # use the default sort
            {"bldr1": 777, "bldr2": 999, "bldr3": 888},
            {"bldr1": 10, "bldr2": 10, "bldr3": 10},
            ['bldr1', 'bldr3', 'bldr2'],
        )

    def test_sortBuilders_custom(self):
        def prioritizeBuilders(master, builders):
            self.assertIdentical(master, self.master)
            return sorted(builders, key=lambda b: b.name)

        return self.do_test_sortBuilders(
            prioritizeBuilders,
            {"bldr1": 1, "bldr2": 1, "bldr3": 1},
            {"bldr1": 10, "bldr2": 15, "bldr3": 5},
            ['bldr1', 'bldr2', 'bldr3'],
        )

    def test_sortBuilders_custom_async(self):
        def prioritizeBuilders(master, builders):
            self.assertIdentical(master, self.master)
            return defer.succeed(sorted(builders, key=lambda b: b.name))

        return self.do_test_sortBuilders(
            prioritizeBuilders,
            {"bldr1": 1, "bldr2": 1, "bldr3": 1},
            {"bldr1": 10, "bldr2": 15, "bldr3": 5},
            ['bldr1', 'bldr2', 'bldr3'],
        )

    @defer.inlineCallbacks
    def test_sortBuilders_custom_exception(self):
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(['x', 'y'])

        def fail(m, b):
            raise RuntimeError("oh noes")

        self.master.config.prioritizeBuilders = fail

        # expect to get the builders back in the same order in the event of an
        # exception
        result = yield self.brd._sortBuilders(['y', 'x'])

        self.assertEqual(result, ['y', 'x'])

        # and expect the exception to be logged
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def test_stopService(self):
        # check that stopService waits for a builder run to complete, but does not
        # allow a subsequent run to start
        self.useMock_maybeStartBuildsOnBuilder()
        self.addBuilders(['A', 'B'])

        oldMSBOB = self.brd._maybeStartBuildsOnBuilder

        def maybeStartBuildsOnBuilder(bldr):
            d = oldMSBOB(bldr)

            stop_d = self.brd.stopService()
            stop_d.addCallback(lambda _: self.maybeStartBuildsOnBuilder_calls.append('(stopped)'))

            d.addCallback(lambda _: self.maybeStartBuildsOnBuilder_calls.append('finished'))
            return d

        self.brd._maybeStartBuildsOnBuilder = maybeStartBuildsOnBuilder

        # start both builds; A should start and complete *before* the service stops,
        # and B should not run.
        yield self.brd.maybeStartBuildsOn(['A', 'B'])

        yield self.brd._waitForFinish()

        self.assertEqual(self.maybeStartBuildsOnBuilder_calls, ['A', 'finished', '(stopped)'])


class TestMaybeStartBuilds(TestBRDBase):
    @defer.inlineCallbacks
    def setUp(self):
        yield super().setUp()

        self.startedBuilds = []

        self.bldr = yield self.createBuilder('A', builderid=77)
        self.builders['A'] = self.bldr

    def assertBuildsStarted(self, exp):
        # munge builds_started into (worker, [brids])
        builds_started = [
            (worker, [br.id for br in breqs]) for (worker, breqs) in self.startedBuilds
        ]
        self.assertEqual(builds_started, exp)

    # _maybeStartBuildsOnBuilder

    @async_to_deferred
    async def do_test_maybeStartBuildsOnBuilder(self, rows=None, exp_claims=None, exp_builds=None):
        rows = rows or []
        exp_claims = exp_claims or []
        exp_builds = exp_builds or []
        await self.master.db.insert_test_data(rows)

        await self.brd._maybeStartBuildsOnBuilder(self.bldr)

        await self.assert_claims(exp_claims)
        self.assertBuildsStarted(exp_builds)

    @defer.inlineCallbacks
    def test_no_buildrequests(self):
        self.addWorkers({'test-worker11': 1})
        yield self.do_test_maybeStartBuildsOnBuilder(exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_no_workerforbuilders(self):
        rows = [
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            fakedb.Builder(id=78, name='bldr'),
            fakedb.Buildset(id=10),
            fakedb.BuildRequest(id=11, buildsetid=10, builderid=78),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_limited_by_workers(self):
        self.addWorkers({'test-worker1': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[10], exp_builds=[('test-worker1', [10])]
        )

    @defer.inlineCallbacks
    def test_sorted_by_submit_time(self):
        # same as "limited_by_workers" but with rows swapped
        self.addWorkers({'test-worker1': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[10], exp_builds=[('test-worker1', [10])]
        )

    @defer.inlineCallbacks
    def test_limited_by_available_workers(self):
        self.addWorkers({'test-worker1': 0, 'test-worker2': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[10], exp_builds=[('test-worker2', [10])]
        )

    @defer.inlineCallbacks
    def test_slow_db(self):
        # test what happens if the "getBuildRequests" fetch takes a "long time"
        self.addWorkers({'test-worker1': 1})

        # wrap to simulate a "long" db access
        old_getBuildRequests = self.master.db.buildrequests.getBuildRequests

        def longGetBuildRequests(*args, **kwargs):
            res_d = old_getBuildRequests(*args, **kwargs)
            long_d = defer.Deferred()
            long_d.addCallback(lambda _: res_d)
            self.reactor.callLater(0, long_d.callback, None)
            return long_d

        self.master.db.buildrequests.getBuildRequests = longGetBuildRequests

        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[10], exp_builds=[('test-worker1', [10])]
        )

    @defer.inlineCallbacks
    def test_limited_by_canStartBuild(self):
        """Set the 'canStartBuild' value in the config to something
        that limits the possible options."""

        self.bldr.config.nextWorker = nth_worker(-1)

        pairs_tested = []

        def _canStartBuild(worker, breq):
            result = (worker.name, breq.id)
            pairs_tested.append(result)
            allowed = [
                ("test-worker1", 10),
                ("test-worker3", 11),
            ]
            return result in allowed

        self.bldr.config.canStartBuild = _canStartBuild

        self.addWorkers({'test-worker1': 1, 'test-worker2': 1, 'test-worker3': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
            fakedb.BuildRequest(id=12, buildsetid=11, builderid=77, submitted_at=140000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows,
            exp_claims=[10, 11],
            exp_builds=[('test-worker1', [10]), ('test-worker3', [11])],
        )

        # we expect brids in order (10-11-12),
        # with each searched in reverse order of workers (3-2-1) available (due
        # to nth_worker(-1))
        self.assertEqual(
            pairs_tested,
            [
                ('test-worker3', 10),
                ('test-worker2', 10),
                ('test-worker1', 10),
                ('test-worker3', 11),
                ('test-worker2', 12),
            ],
        )

    @defer.inlineCallbacks
    def test_limited_by_canStartBuild_deferreds(self):
        # Another variant that returns Deferred types,
        self.bldr.config.nextWorker = nth_worker(-1)

        pairs_tested = []

        def _canStartBuild(worker, breq):
            result = (worker.name, breq.id)
            pairs_tested.append(result)
            allowed = [
                ("test-worker1", 10),
                ("test-worker3", 11),
            ]
            return defer.succeed(result in allowed)

        self.bldr.config.canStartBuild = _canStartBuild

        self.addWorkers({'test-worker1': 1, 'test-worker2': 1, 'test-worker3': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
            fakedb.BuildRequest(id=12, buildsetid=11, builderid=77, submitted_at=140000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows,
            exp_claims=[10, 11],
            exp_builds=[('test-worker1', [10]), ('test-worker3', [11])],
        )

        # we expect brids in order (10-11-12),
        # with worker2 unable to pair
        self.assertEqual(
            pairs_tested,
            [
                ('test-worker3', 10),
                ('test-worker2', 10),
                ('test-worker1', 10),
                ('test-worker3', 11),
                ('test-worker2', 12),
            ],
        )

    @defer.inlineCallbacks
    def test_unlimited(self):
        self.bldr.config.nextWorker = nth_worker(-1)
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows,
            exp_claims=[10, 11],
            exp_builds=[('test-worker2', [10]), ('test-worker1', [11])],
        )

    @defer.inlineCallbacks
    def test_bldr_maybeStartBuild_fails_always(self):
        self.bldr.config.nextWorker = nth_worker(-1)
        # the builder fails to start the build; we'll see that the build
        # was requested, but the brids will get claimed again

        def maybeStartBuild(worker, builds):
            self.startedBuilds.append((worker.name, builds))
            return defer.succeed(False)

        self.bldr.maybeStartBuild = maybeStartBuild

        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows,
            # claimed again so none taken!
            exp_claims=[],
            exp_builds=[('test-worker2', [10]), ('test-worker1', [11])],
        )

    @async_to_deferred
    async def test_bldr_maybeStartBuild_fails_once(self):
        self.bldr.config.nextWorker = nth_worker(-1)
        # the builder fails to start the build; we'll see that the build
        # was requested, but the brids will get claimed again
        start_build_results = [False, True, True]

        def maybeStartBuild(worker, builds):
            self.startedBuilds.append((worker.name, builds))
            return defer.succeed(start_build_results.pop(0))

        self.bldr.maybeStartBuild = maybeStartBuild

        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [
            *self.base_rows,
            fakedb.BuildRequest(id=10, buildsetid=11, builderid=77, submitted_at=130000),
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]

        await self.master.db.insert_test_data(rows)

        # first time around, only #11 stays claimed
        await self.brd._maybeStartBuildsOnBuilder(self.bldr)
        await self.assert_claims([11])  # claimed again so none taken!
        self.assertBuildsStarted([('test-worker2', [10]), ('test-worker1', [11])])

        # second time around the #10 will pass, adding another request and it
        # is claimed
        await self.brd._maybeStartBuildsOnBuilder(self.bldr)
        await self.assert_claims([10, 11])
        self.assertBuildsStarted([
            ('test-worker2', [10]),
            ('test-worker1', [11]),
            ('test-worker2', [10]),
        ])

    @defer.inlineCallbacks
    def test_limited_by_requests(self):
        self.bldr.config.nextWorker = nth_worker(1)
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[11], exp_builds=[('test-worker2', [11])]
        )

    @defer.inlineCallbacks
    def test_nextWorker_None(self):
        self.bldr.config.nextWorker = lambda _1, _2, _3: defer.succeed(None)
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        yield self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_nextWorker_bogus(self):
        self.bldr.config.nextWorker = lambda _1, _2, _3: defer.succeed(mock.Mock())
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        yield self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_nextBuild_None(self):
        self.bldr.config.nextBuild = lambda _1, _2: defer.succeed(None)
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        yield self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_nextBuild_bogus(self):
        self.bldr.config.nextBuild = lambda _1, _2: mock.Mock()
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        yield self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])

    @defer.inlineCallbacks
    def test_nextBuild_fails(self):
        def nextBuildRaises(*args):
            raise RuntimeError("xx")

        self.bldr.config.nextBuild = nextBuildRaises
        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [*self.base_rows, fakedb.BuildRequest(id=11, buildsetid=11, builderid=77)]
        result = self.do_test_maybeStartBuildsOnBuilder(rows=rows, exp_claims=[], exp_builds=[])
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))
        yield result

    # check concurrency edge cases
    @defer.inlineCallbacks
    def test_claim_race(self):
        self.bldr.config.nextWorker = nth_worker(0)
        # fake a race condition on the buildrequests table
        old_claimBuildRequests = self.master.db.buildrequests.claimBuildRequests

        def claimBuildRequests(brids, claimed_at=None):
            # first, ensure this only happens the first time
            self.master.db.buildrequests.claimBuildRequests = old_claimBuildRequests
            # claim brid 10 for some other master
            assert 10 in brids
            self.master.db.buildrequests._claim_buildrequests_for_master(
                [10], 136000, 9999
            )  # some other masterid
            # ..and fail
            return defer.fail(buildrequests.AlreadyClaimedError())

        self.master.db.buildrequests.claimBuildRequests = claimBuildRequests

        self.addWorkers({'test-worker1': 1, 'test-worker2': 1})
        rows = [
            *self.base_rows,
            fakedb.Master(id=9999),
            fakedb.BuildRequest(
                id=10, buildsetid=11, builderid=77, submitted_at=130000
            ),  # will turn out to be claimed!
            fakedb.BuildRequest(id=11, buildsetid=11, builderid=77, submitted_at=135000),
        ]
        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=[11], exp_builds=[('test-worker1', [11])]
        )

    # nextWorker
    @defer.inlineCallbacks
    def do_test_nextWorker(self, nextWorker, global_select_next_worker, exp_choice=None):
        if global_select_next_worker:
            self.master.config.select_next_worker = nextWorker
            builder_config = config.BuilderConfig(
                name='bldrconf',
                workernames=['wk1', 'wk2'],
                builddir='bdir',
                factory=factory.BuildFactory(),
            )
        else:
            builder_config = config.BuilderConfig(
                name='bldrconf',
                workernames=['wk1', 'wk2'],
                builddir='bdir',
                factory=factory.BuildFactory(),
                nextWorker=nextWorker,
            )

        self.bldr = yield self.createBuilder('B', builderid=78, builder_config=builder_config)
        for i in range(4):
            self.addWorkers({f'test-worker{i}': 1})

        rows = [
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
            fakedb.SourceStamp(id=21),
            fakedb.Builder(id=78, name='B'),
            fakedb.Buildset(id=12, reason='because'),
            fakedb.BuildsetSourceStamp(sourcestampid=21, buildsetid=12),
            fakedb.BuildRequest(id=12, buildsetid=12, builderid=78),
        ]

        if exp_choice is None:
            exp_claims = []
            exp_builds = []
        else:
            exp_claims = [12]
            exp_builds = [(f'test-worker{exp_choice}', [12])]

        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=exp_claims, exp_builds=exp_builds
        )

    @parameterized.expand([True, False])
    def test_nextWorker_gets_buildrequest(self, global_select_next_worker):
        def nextWorker(bldr, lst, br=None):
            self.assertNotEqual(br, None)

        return self.do_test_nextWorker(
            nextWorker, global_select_next_worker=global_select_next_worker
        )

    @parameterized.expand([True, False])
    def test_nextWorker_default(self, global_select_next_worker):
        self.patch(random, 'choice', nth_worker(2))
        return self.do_test_nextWorker(
            None, exp_choice=2, global_select_next_worker=global_select_next_worker
        )

    @parameterized.expand([True, False])
    def test_nextWorker_simple(self, global_select_next_worker):
        def nextWorker(bldr, lst, br=None):
            self.assertIdentical(bldr, self.bldr)
            return lst[1]

        return self.do_test_nextWorker(
            nextWorker, global_select_next_worker=global_select_next_worker, exp_choice=1
        )

    @parameterized.expand([True, False])
    def test_nextWorker_deferred(self, global_select_next_worker):
        def nextWorker(bldr, lst, br=None):
            self.assertIdentical(bldr, self.bldr)
            return defer.succeed(lst[1])

        return self.do_test_nextWorker(
            nextWorker, global_select_next_worker=global_select_next_worker, exp_choice=1
        )

    @parameterized.expand([True, False])
    @defer.inlineCallbacks
    def test_nextWorker_exception(self, global_select_next_worker):
        def nextWorker(bldr, lst, br=None):
            raise RuntimeError("")

        yield self.do_test_nextWorker(
            nextWorker, global_select_next_worker=global_select_next_worker
        )
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))

    @parameterized.expand([True, False])
    @defer.inlineCallbacks
    def test_nextWorker_failure(self, global_select_next_worker):
        def nextWorker(bldr, lst, br=None):
            return defer.fail(failure.Failure(RuntimeError()))

        yield self.do_test_nextWorker(
            nextWorker, global_select_next_worker=global_select_next_worker
        )
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))

    # _nextBuild

    @defer.inlineCallbacks
    def do_test_nextBuild(self, nextBuild, exp_choice=None):
        self.bldr.config.nextWorker = nth_worker(-1)
        self.bldr.config.nextBuild = nextBuild

        rows = self.make_workers(4)

        exp_claims = []
        exp_builds = []
        if exp_choice is not None:
            worker = 3
            for choice in exp_choice:
                exp_claims.append(choice)
                exp_builds.append((f'test-worker{worker}', [choice]))
                worker = worker - 1

        yield self.do_test_maybeStartBuildsOnBuilder(
            rows=rows, exp_claims=sorted(exp_claims), exp_builds=exp_builds
        )

    def test_nextBuild_default(self):
        "default chooses the first in the list, which should be the earliest"
        return self.do_test_nextBuild(None, exp_choice=[10, 11, 12, 13])

    def test_nextBuild_simple(self):
        def nextBuild(bldr, lst):
            self.assertIdentical(bldr, self.bldr)
            return lst[-1]

        return self.do_test_nextBuild(nextBuild, exp_choice=[13, 12, 11, 10])

    def test_nextBuild_deferred(self):
        def nextBuild(bldr, lst):
            self.assertIdentical(bldr, self.bldr)
            return defer.succeed(lst[-1])

        return self.do_test_nextBuild(nextBuild, exp_choice=[13, 12, 11, 10])

    def test_nextBuild_exception(self):
        def nextBuild(bldr, lst):
            raise RuntimeError("")

        result = self.do_test_nextBuild(nextBuild)
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))
        return result

    def test_nextBuild_failure(self):
        def nextBuild(bldr, lst):
            return defer.fail(failure.Failure(RuntimeError()))

        result = self.do_test_nextBuild(nextBuild)
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))
        return result



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# Method to add build step taken from here
# https://seasonofcode.com/posts/how-to-add-custom-build-steps-and-commands-to-setuppy.html
import datetime
import logging
import os
import re
import shutil
import subprocess
import sys
from subprocess import PIPE
from subprocess import STDOUT
from subprocess import Popen

import setuptools.command.build_py
import setuptools.command.egg_info
from setuptools import Command
from setuptools import setup

old_listdir = os.listdir


def listdir(path):
    # patch listdir to avoid looking into node_modules
    l = old_listdir(path)
    if "node_modules" in l:
        l.remove("node_modules")
    return l


os.listdir = listdir


def check_output(cmd, shell):
    """Version of check_output which does not throw error"""
    popen = subprocess.Popen(cmd, shell=shell, stdout=subprocess.PIPE)
    out = popen.communicate()[0].strip()
    if not isinstance(out, str):
        out = out.decode(sys.stdout.encoding)
    return out


def gitDescribeToPep440(version):
    # git describe produce version in the form: v0.9.8-20-gf0f45ca
    # where 20 is the number of commit since last release, and gf0f45ca is the short commit id preceded by 'g'
    # we parse this a transform into a pep440 release version 0.9.9.dev20 (increment last digit and add dev before 20)

    VERSION_MATCH = re.compile(
        r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(\.post(?P<post>\d+))?(-(?P<dev>\d+))?(-g(?P<commit>.+))?'
    )
    v = VERSION_MATCH.search(version)
    if v:
        major = int(v.group('major'))
        minor = int(v.group('minor'))
        patch = int(v.group('patch'))
        if v.group('dev'):
            patch += 1
            dev = int(v.group('dev'))
            return f"{major}.{minor}.{patch}.dev{dev}"
        if v.group('post'):
            return "{}.{}.{}.post{}".format(major, minor, patch, v.group('post'))
        return f"{major}.{minor}.{patch}"

    return v


def mTimeVersion(init_file):
    cwd = os.path.dirname(os.path.abspath(init_file))
    m = 0
    for root, dirs, files in os.walk(cwd):
        for f in files:
            m = max(os.path.getmtime(os.path.join(root, f)), m)
    d = datetime.datetime.fromtimestamp(m, datetime.timezone.utc)
    return d.strftime("%Y.%m.%d")


def getVersionFromArchiveId(git_archive_id='1747081323 %(describe:abbrev=10)'):
    """Extract the tag if a source is from git archive.

    When source is exported via `git archive`, the git_archive_id init value is modified
    and placeholders are expanded to the "archived" revision:

        %ct: committer date, UNIX timestamp
        %(describe:abbrev=10): git-describe output, always abbreviating to 10 characters of commit ID.
                               e.g. v3.10.0-850-g5bf957f89

    See man gitattributes(5) and git-log(1) (PRETTY FORMATS) for more details.
    """
    # mangle the magic string to make sure it is not replaced by git archive
    if not git_archive_id.startswith('$For' + 'mat:'):
        # source was modified by git archive, try to parse the version from
        # the value of git_archive_id

        tstamp, _, describe_output = git_archive_id.strip().partition(' ')
        if describe_output:
            # archived revision is tagged, use the tag
            return gitDescribeToPep440(describe_output)

        # archived revision is not tagged, use the commit date
        d = datetime.datetime.fromtimestamp(int(tstamp), datetime.timezone.utc)
        return d.strftime('%Y.%m.%d')
    return None


def getVersion(init_file):
    """
    Return BUILDBOT_VERSION environment variable, content of VERSION file, git
    tag or '0.0.0' meaning we could not find the version, but the output still has to be valid
    """

    try:
        return os.environ['BUILDBOT_VERSION']
    except KeyError:
        pass

    try:
        cwd = os.path.dirname(os.path.abspath(init_file))
        fn = os.path.join(cwd, 'VERSION')
        with open(fn) as f:
            return f.read().strip()
    except OSError:
        pass

    version = getVersionFromArchiveId()
    if version is not None:
        return version

    try:
        p = Popen(['git', 'describe', '--tags', '--always'], stdout=PIPE, stderr=STDOUT, cwd=cwd)
        out = p.communicate()[0]

        if (not p.returncode) and out:
            v = gitDescribeToPep440(str(out))
            if v:
                return v
    except OSError:
        pass

    try:
        # if we really can't find the version, we use the date of modification of the most recent file
        # docker hub builds cannot use git describe
        return mTimeVersion(init_file)
    except Exception:
        # bummer. lets report something
        return "0.0.0"


# JS build strategy:
#
# Obviously, building javascript with setuptools is not really something supported initially
#
# The goal of this hack are:
# - override the distutils command to insert our js build
# - has very small setup.py
#
# from buildbot_pkg import setup_www
#
# setup_www(
#   ...
#    packages=["buildbot_myplugin"]
# )
#
# We need to override the first command done, so that source tree is populated very soon,
# as well as version is found from git tree or "VERSION" file
#
# This supports following setup.py commands:
#
# - develop, via egg_info
# - install, via egg_info
# - sdist, via egg_info
# - bdist_wheel, via build
# This is why we override both egg_info and build, and the first run build
# the js.


class BuildJsCommand(Command):
    """A custom command to run JS build."""

    description = 'run JS build'
    already_run = False

    def initialize_options(self):
        """Set default values for options."""

    def finalize_options(self):
        """Post-process options."""

    def run(self):
        """Run command."""
        if self.already_run:
            return

        if os.path.isdir('build'):
            shutil.rmtree('build')

        package = self.distribution.packages[0]
        if os.path.exists("package.json"):
            shell = bool(os.name == 'nt')

            yarn_program = None
            for program in ["yarnpkg", "yarn"]:
                try:
                    yarn_version = check_output([program, "--version"], shell=shell)
                    if yarn_version != "":
                        yarn_program = program
                        break
                except subprocess.CalledProcessError:
                    pass

            assert yarn_program is not None, "need nodejs and yarn installed in current PATH"

            commands = [
                [yarn_program, 'install', '--pure-lockfile'],
                [yarn_program, 'run', 'build'],
            ]

            for command in commands:
                logging.info('Running command: {}'.format(str(" ".join(command))))
                try:
                    subprocess.check_call(command, shell=shell)
                except subprocess.CalledProcessError as e:
                    raise Exception(
                        f"Exception = {e} command was called in directory = {os.getcwd()}"
                    ) from e

        self.copy_tree(
            os.path.join(package, 'static'), os.path.join("build", "lib", package, "static")
        )

        assert self.distribution.metadata.version is not None, "version is not set"
        with open(os.path.join("build", "lib", package, "VERSION"), "w") as f:
            f.write(self.distribution.metadata.version)

        with open(os.path.join(package, "VERSION"), "w") as f:
            f.write(self.distribution.metadata.version)

        self.already_run = True


class BuildPyCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    def run(self):
        self.run_command('build_js')
        super().run()


class EggInfoCommand(setuptools.command.egg_info.egg_info):
    """Custom egginfo command."""

    def run(self):
        self.run_command('build_js')
        super().run()


def setup_www_plugin(**kw):
    package = kw['packages'][0]
    if 'version' not in kw:
        kw['version'] = getVersion(os.path.join(package, "__init__.py"))

    setup(
        cmdclass=dict(egg_info=EggInfoCommand, build_py=BuildPyCommand, build_js=BuildJsCommand),
        **kw,
    )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log

from buildbot.util import deferwaiter
from buildbot.util import service


class MQBase(service.AsyncService):
    name = 'mq-implementation'

    def __init__(self):
        super().__init__()
        self._deferwaiter = deferwaiter.DeferWaiter()

    @defer.inlineCallbacks
    def stopService(self):
        yield self._deferwaiter.wait()
        yield super().stopService()

    @defer.inlineCallbacks
    def waitUntilEvent(self, filter, check_callback):
        d = defer.Deferred()
        buildCompleteConsumer = yield self.startConsuming(
            lambda key, value: d.callback((key, value)), filter
        )
        check = yield check_callback()
        # we only wait if the check callback return true
        if not check:
            res = yield d
        else:
            res = None
        yield buildCompleteConsumer.stopConsuming()
        return res

    def invokeQref(self, qref, routingKey, data):
        self._deferwaiter.add(qref.invoke(routingKey, data))


class QueueRef:
    __slots__ = ['callback']

    def __init__(self, callback):
        self.callback = callback

    def invoke(self, routing_key, data):
        # Potentially returns a Deferred
        if not self.callback:
            return None

        try:
            x = self.callback(routing_key, data)
        except Exception as e:
            log.err(e, f'while invoking {self.callback!r}')
            return None
        if isinstance(x, defer.Deferred):
            x.addErrback(log.err, f'while invoking {self.callback!r}')
        return x

    def stopConsuming(self):
        # This method may return a Deferred.
        # subclasses should set self.callback to None in this method.
        raise NotImplementedError



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import resultspec
from buildbot.db import builds
from buildbot.db.builds import BuildModel
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import UTC
from buildbot.util import epoch2datetime

TIME1 = 1304262222
TIME2 = 1304262223
TIME3 = 1304262224
TIME4 = 1304262235
CREATED_AT = 927845299


class Tests(TestReactorMixin, unittest.TestCase):
    # common sample data

    backgroundData = [
        fakedb.Buildset(id=20),
        fakedb.Builder(id=77, name="b1"),
        fakedb.Builder(id=88, name="b2"),
        fakedb.BuildRequest(id=40, buildsetid=20, builderid=77),
        fakedb.BuildRequest(id=41, buildsetid=20, builderid=77),
        fakedb.BuildRequest(id=42, buildsetid=20, builderid=88),
        fakedb.Master(id=88),
        fakedb.Master(id=89, name="bar"),
        fakedb.Worker(id=13, name='wrk'),
        fakedb.Worker(id=12, name='sl2'),
    ]

    threeBuilds = [
        fakedb.Build(
            id=50,
            buildrequestid=42,
            number=5,
            masterid=88,
            builderid=77,
            workerid=13,
            state_string="build 5",
            started_at=TIME1,
        ),
        fakedb.Build(
            id=51,
            buildrequestid=41,
            number=6,
            masterid=88,
            builderid=88,
            workerid=13,
            state_string="build 6",
            started_at=TIME2,
        ),
        fakedb.Build(
            id=52,
            buildrequestid=42,
            number=7,
            masterid=88,
            builderid=77,
            workerid=12,
            state_string="build 7",
            started_at=TIME3,
            complete_at=TIME4,
            results=5,
        ),
    ]

    threeBdicts = {
        50: builds.BuildModel(
            id=50,
            buildrequestid=42,
            builderid=77,
            masterid=88,
            number=5,
            workerid=13,
            started_at=epoch2datetime(TIME1),
            complete_at=None,
            locks_duration_s=0,
            state_string="build 5",
            results=None,
        ),
        51: builds.BuildModel(
            id=51,
            buildrequestid=41,
            builderid=88,
            masterid=88,
            number=6,
            workerid=13,
            started_at=epoch2datetime(TIME2),
            complete_at=None,
            locks_duration_s=0,
            state_string="build 6",
            results=None,
        ),
        52: builds.BuildModel(
            id=52,
            buildrequestid=42,
            builderid=77,
            masterid=88,
            number=7,
            workerid=12,
            started_at=epoch2datetime(TIME3),
            complete_at=epoch2datetime(TIME4),
            locks_duration_s=0,
            state_string="build 7",
            results=5,
        ),
    }

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_getBuild(self):
        yield self.db.insert_test_data([*self.backgroundData, self.threeBuilds[0]])
        bdict = yield self.db.builds.getBuild(50)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                id=50,
                number=5,
                buildrequestid=42,
                masterid=88,
                builderid=77,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=0,
                state_string='build 5',
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuild_missing(self):
        bdict = yield self.db.builds.getBuild(50)
        self.assertEqual(bdict, None)

    @defer.inlineCallbacks
    def test_getBuildByNumber(self):
        yield self.db.insert_test_data([*self.backgroundData, self.threeBuilds[0]])
        bdict = yield self.db.builds.getBuildByNumber(builderid=77, number=5)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(bdict.id, 50)

    @defer.inlineCallbacks
    def test_getBuilds(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds()
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            sorted(bdicts, key=lambda bd: bd.id),
            [self.threeBdicts[50], self.threeBdicts[51], self.threeBdicts[52]],
        )

    @defer.inlineCallbacks
    def test_getBuilds_builderid(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(builderid=88)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[51]])

    @defer.inlineCallbacks
    def test_getBuilds_buildrequestid(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(buildrequestid=42)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[50], self.threeBdicts[52]]
        )

    @defer.inlineCallbacks
    def test_getBuilds_workerid(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(workerid=13)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[50], self.threeBdicts[51]]
        )

    @defer.inlineCallbacks
    def do_test_getBuildsForChange(self, rows, changeid, expected):
        yield self.db.insert_test_data(rows)

        builds = yield self.db.builds.getBuildsForChange(changeid)

        self.assertEqual(sorted(builds), sorted(expected))

    def test_getBuildsForChange_OneCodebase(self):
        rows = [
            fakedb.Master(id=88, name="bar"),
            fakedb.Worker(id=13, name='one'),
            fakedb.Builder(id=77, name='A'),
            fakedb.SourceStamp(id=234, created_at=CREATED_AT, revision="aaa"),
            fakedb.Change(changeid=14, codebase='A', sourcestampid=234),
            fakedb.Buildset(id=30, reason='foo', submitted_at=1300305712, results=1),
            fakedb.BuildsetSourceStamp(sourcestampid=234, buildsetid=30),
            fakedb.BuildRequest(
                id=19,
                buildsetid=30,
                builderid=77,
                priority=13,
                submitted_at=1300305712,
                results=1,
                complete=0,
                complete_at=None,
            ),
            fakedb.Build(
                id=50,
                buildrequestid=19,
                number=5,
                masterid=88,
                builderid=77,
                state_string="test",
                workerid=13,
                started_at=1304262222,
                results=1,
            ),
        ]

        expected = [
            builds.BuildModel(
                id=50,
                number=5,
                builderid=77,
                buildrequestid=19,
                workerid=13,
                masterid=88,
                started_at=epoch2datetime(1304262222),
                complete_at=None,
                locks_duration_s=0,
                state_string='test',
                results=1,
            ),
        ]

        return self.do_test_getBuildsForChange(rows, 14, expected)

    @defer.inlineCallbacks
    def test_getBuilds_complete(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(complete=True)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[52]])

    @defer.inlineCallbacks
    def test_addBuild_first(self):
        self.reactor.advance(TIME1)
        yield self.db.insert_test_data(self.backgroundData)
        id, number = yield self.db.builds.addBuild(
            builderid=77, buildrequestid=41, workerid=13, masterid=88, state_string='test test2'
        )
        bdict = yield self.db.builds.getBuild(id)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                buildrequestid=41,
                builderid=77,
                id=id,
                masterid=88,
                number=number,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=0,
                state_string="test test2",
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_addBuild_existing(self):
        self.reactor.advance(TIME1)
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Build(number=10, buildrequestid=41, builderid=77, masterid=88, workerid=13),
        ])
        id, number = yield self.db.builds.addBuild(
            builderid=77, buildrequestid=41, workerid=13, masterid=88, state_string='test test2'
        )
        bdict = yield self.db.builds.getBuild(id)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(number, 11)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                buildrequestid=41,
                builderid=77,
                id=id,
                masterid=88,
                number=number,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=0,
                state_string="test test2",
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_setBuildStateString(self):
        yield self.db.insert_test_data([*self.backgroundData, self.threeBuilds[0]])
        yield self.db.builds.setBuildStateString(buildid=50, state_string='test test2')
        bdict = yield self.db.builds.getBuild(50)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                id=50,
                number=5,
                buildrequestid=42,
                masterid=88,
                builderid=77,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=0,
                state_string='test test2',
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_add_build_locks_duration(self):
        yield self.db.insert_test_data([*self.backgroundData, self.threeBuilds[0]])
        yield self.db.builds.add_build_locks_duration(buildid=50, duration_s=12)
        bdict = yield self.db.builds.getBuild(50)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                id=50,
                number=5,
                buildrequestid=42,
                masterid=88,
                builderid=77,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=12,
                state_string="build 5",
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_finishBuild(self):
        self.reactor.advance(TIME4)
        yield self.db.insert_test_data([*self.backgroundData, self.threeBuilds[0]])
        yield self.db.builds.finishBuild(buildid=50, results=7)
        bdict = yield self.db.builds.getBuild(50)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                id=50,
                number=5,
                buildrequestid=42,
                masterid=88,
                builderid=77,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=epoch2datetime(TIME4),
                locks_duration_s=0,
                state_string='build 5',
                results=7,
            ),
        )

    @defer.inlineCallbacks
    def testgetBuildPropertiesEmpty(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        for buildid in (50, 51, 52):
            props = yield self.db.builds.getBuildProperties(buildid)
            self.assertEqual(0, len(props))

    @defer.inlineCallbacks
    def test_testgetBuildProperties_resultSpecFilter(self):
        rs = resultspec.ResultSpec(filters=[resultspec.Filter('name', 'eq', ["prop", "prop2"])])
        rs.fieldMapping = {'name': 'build_properties.name'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        yield self.db.builds.setBuildProperty(50, 'prop', 42, 'test')
        yield self.db.builds.setBuildProperty(50, 'prop2', 43, 'test')
        yield self.db.builds.setBuildProperty(50, 'prop3', 44, 'test')
        props = yield self.db.builds.getBuildProperties(50, resultSpec=rs)
        self.assertEqual(props, {'prop': (42, 'test'), 'prop2': (43, 'test')})

        rs = resultspec.ResultSpec(filters=[resultspec.Filter('name', 'eq', ["prop"])])
        rs.fieldMapping = {'name': 'build_properties.name'}
        props = yield self.db.builds.getBuildProperties(50, resultSpec=rs)
        self.assertEqual(
            props,
            {
                'prop': (42, 'test'),
            },
        )

    @defer.inlineCallbacks
    def testsetandgetProperties(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        yield self.db.builds.setBuildProperty(50, 'prop', 42, 'test')
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {'prop': (42, 'test')})

    @defer.inlineCallbacks
    def testsetgetsetProperties(self):
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {})
        yield self.db.builds.setBuildProperty(50, 'prop', 42, 'test')
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {'prop': (42, 'test')})
        # set a new value
        yield self.db.builds.setBuildProperty(50, 'prop', 45, 'test')
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {'prop': (45, 'test')})
        # set a new source
        yield self.db.builds.setBuildProperty(50, 'prop', 45, 'test_source')
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {'prop': (45, 'test_source')})
        # set the same
        yield self.db.builds.setBuildProperty(50, 'prop', 45, 'test_source')
        props = yield self.db.builds.getBuildProperties(50)
        self.assertEqual(props, {'prop': (45, 'test_source')})

    @defer.inlineCallbacks
    def test_addBuild_existing_race(self):
        self.reactor.advance(TIME1)
        yield self.db.insert_test_data(self.backgroundData)

        # add new builds at *just* the wrong time, repeatedly
        numbers = list(range(1, 8))

        def raceHook(conn):
            if not numbers:
                return
            conn.execute(
                self.db.model.builds.insert(),
                {
                    "number": numbers.pop(0),
                    "buildrequestid": 41,
                    "masterid": 88,
                    "workerid": 13,
                    "builderid": 77,
                    "started_at": TIME1,
                    "locks_duration_s": 0,
                    "state_string": "hi",
                },
            )
            conn.commit()

        id, number = yield self.db.builds.addBuild(
            builderid=77,
            buildrequestid=41,
            workerid=13,
            masterid=88,
            state_string='test test2',
            _race_hook=raceHook,
        )
        bdict = yield self.db.builds.getBuild(id)
        self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(number, 8)
        self.assertEqual(
            bdict,
            builds.BuildModel(
                buildrequestid=41,
                builderid=77,
                id=id,
                masterid=88,
                number=number,
                workerid=13,
                started_at=epoch2datetime(TIME1),
                complete_at=None,
                locks_duration_s=0,
                state_string="test test2",
                results=None,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecFilter(self):
        rs = resultspec.ResultSpec(filters=[resultspec.Filter('complete_at', 'ne', [None])])
        rs.fieldMapping = {'complete_at': 'builds.complete_at'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[52]])

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecOrder(self):
        rs = resultspec.ResultSpec(order=['-started_at'])
        rs.fieldMapping = {'started_at': 'builds.started_at'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)

        # applying the spec in the db layer should have emptied the order in
        # resultSpec
        self.assertEqual(rs.order, None)
        # assert applying the same order at the data layer will give the same
        # results
        rs = resultspec.ResultSpec(order=['-started_at'])
        ordered_bdicts = rs.apply(bdicts)
        self.assertEqual(ordered_bdicts, bdicts)

        # assert applying an opposite order at the data layer will give different
        # results
        rs = resultspec.ResultSpec(order=['started_at'])
        ordered_bdicts = rs.apply(bdicts)
        self.assertNotEqual(ordered_bdicts, bdicts)

    @defer.inlineCallbacks
    def test_getBuilds_limit(self):
        rs = resultspec.ResultSpec(order=['-started_at'], limit=1, offset=2)
        rs.fieldMapping = {'started_at': 'builds.started_at'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        # applying the spec in the db layer should have emptied the limit and
        # offset in resultSpec
        self.assertEqual(rs.limit, None)
        self.assertEqual(rs.offset, None)

        # assert applying the same filter at the data layer will give the same
        # results
        rs = resultspec.ResultSpec(order=['-started_at'], limit=1, offset=2)
        bdicts2 = yield self.db.builds.getBuilds()
        ordered_bdicts = rs.apply(bdicts2)
        self.assertEqual(ordered_bdicts, bdicts)

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecFilterEqTwoValues(self):
        rs = resultspec.ResultSpec(filters=[resultspec.Filter('number', 'eq', [6, 7])])
        rs.fieldMapping = {'number': 'builds.number'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[51], self.threeBdicts[52]]
        )

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecFilterNeTwoValues(self):
        rs = resultspec.ResultSpec(filters=[resultspec.Filter('number', 'ne', [6, 7])])
        rs.fieldMapping = {'number': 'builds.number'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[50]])

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecFilterContainsOneValue(self):
        rs = resultspec.ResultSpec(filters=[resultspec.Filter('state_string', 'contains', ['7'])])
        rs.fieldMapping = {'state_string': 'builds.state_string'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[52]])

    @defer.inlineCallbacks
    def test_getBuilds_resultSpecFilterContainsTwoValues(self):
        rs = resultspec.ResultSpec(
            filters=[resultspec.Filter('state_string', 'contains', ['build 5', 'build 6'])]
        )
        rs.fieldMapping = {'state_string': 'builds.state_string'}
        yield self.db.insert_test_data(self.backgroundData + self.threeBuilds)
        bdicts = yield self.db.builds.getBuilds(resultSpec=rs)
        for bdict in bdicts:
            self.assertIsInstance(bdict, builds.BuildModel)
        self.assertEqual(
            sorted(bdicts, key=lambda bd: bd.id), [self.threeBdicts[50], self.threeBdicts[51]]
        )

    @defer.inlineCallbacks
    def test_get_triggered_builds(self):
        yield self.db.insert_test_data(
            self.backgroundData
            + self.threeBuilds
            + [
                fakedb.Buildset(id=1000, parent_buildid=51),
                fakedb.BuildRequest(id=1100, buildsetid=1000, builderid=77),
                fakedb.BuildRequest(id=1101, buildsetid=1000, builderid=77),
                fakedb.Build(id=1200, buildrequestid=1100, masterid=88, builderid=77, workerid=13),
                fakedb.Build(id=1201, buildrequestid=1101, masterid=88, builderid=77, workerid=13),
            ]
        )

        builds = yield self.db.builds.get_triggered_builds(50)
        self.assertEqual(builds, [])

        builds = yield self.db.builds.get_triggered_builds(51)
        self.assertEqual(
            builds,
            [
                BuildModel(
                    id=1200,
                    number=1200,
                    builderid=77,
                    buildrequestid=1100,
                    workerid=13,
                    masterid=88,
                    started_at=datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=UTC),
                    complete_at=None,
                    locks_duration_s=0,
                    state_string='test',
                    results=None,
                ),
                BuildModel(
                    id=1201,
                    number=1201,
                    builderid=77,
                    buildrequestid=1101,
                    workerid=13,
                    masterid=88,
                    started_at=datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=UTC),
                    complete_at=None,
                    locks_duration_s=0,
                    state_string='test',
                    results=None,
                ),
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import builders
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.twisted import async_to_deferred


def builderKey(builder: builders.BuilderModel) -> int:
    return builder.id


class Tests(TestReactorMixin, unittest.TestCase):
    # common sample data

    builder_row = [
        fakedb.Builder(id=7, name="some:builder"),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_updateBuilderInfo(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=123, name="fake_project123"),
            fakedb.Project(id=124, name="fake_project124"),
            fakedb.Builder(id=7, name='some:builder7'),
            fakedb.Builder(id=8, name='some:builder8'),
        ])

        yield self.db.builders.updateBuilderInfo(
            7, 'a string which describe the builder', None, None, 123, ['cat1', 'cat2']
        )
        yield self.db.builders.updateBuilderInfo(
            8, 'a string which describe the builder', None, None, 124, []
        )
        builderdict7 = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict7,
            builders.BuilderModel(
                id=7,
                name='some:builder7',
                tags=["cat1", "cat2"],
                description="a string which describe the builder",
                projectid=123,
            ),
        )

        builderdict8 = yield self.db.builders.getBuilder(8)
        self.assertEqual(
            builderdict8,
            builders.BuilderModel(
                id=8,
                name='some:builder8',
                description="a string which describe the builder",
                projectid=124,
            ),
        )

    @defer.inlineCallbacks
    def test_update_builder_info_tags_case(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=107, name='fake_project'),
            fakedb.Builder(id=7, name='some:builder7', projectid=107),
        ])

        yield self.db.builders.updateBuilderInfo(7, 'builder_desc', None, None, 107, ['Cat', 'cat'])
        builder_dict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builder_dict,
            builders.BuilderModel(
                id=7,
                name='some:builder7',
                tags=['Cat', 'cat'],
                description='builder_desc',
                projectid=107,
            ),
        )

    @defer.inlineCallbacks
    def test_findBuilderId_new(self):
        id = yield self.db.builders.findBuilderId('some:builder')
        builderdict = yield self.db.builders.getBuilder(id)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=id,
                name='some:builder',
            ),
        )

    @defer.inlineCallbacks
    def test_findBuilderId_new_no_autoCreate(self):
        id = yield self.db.builders.findBuilderId('some:builder', autoCreate=False)
        self.assertIsNone(id)

    @defer.inlineCallbacks
    def test_findBuilderId_exists(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7, name='some:builder'),
        ])
        id = yield self.db.builders.findBuilderId('some:builder')
        self.assertEqual(id, 7)

    @defer.inlineCallbacks
    def test_addBuilderMaster(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7),
            fakedb.Master(id=9, name='abc'),
            fakedb.Master(id=10, name='def'),
            fakedb.BuilderMaster(builderid=7, masterid=10),
        ])
        yield self.db.builders.addBuilderMaster(builderid=7, masterid=9)
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=7,
                name='builder-7',
                masterids=[9, 10],
            ),
        )

    @defer.inlineCallbacks
    def test_addBuilderMaster_already_present(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7),
            fakedb.Master(id=9, name='abc'),
            fakedb.Master(id=10, name='def'),
            fakedb.BuilderMaster(builderid=7, masterid=9),
        ])
        yield self.db.builders.addBuilderMaster(builderid=7, masterid=9)
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=7,
                name='builder-7',
                masterids=[9],
            ),
        )

    @defer.inlineCallbacks
    def test_removeBuilderMaster(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7),
            fakedb.Master(id=9),
            fakedb.Master(id=10),
            fakedb.BuilderMaster(builderid=7, masterid=9),
            fakedb.BuilderMaster(builderid=7, masterid=10),
        ])
        yield self.db.builders.removeBuilderMaster(builderid=7, masterid=9)
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=7,
                name='builder-7',
                masterids=[10],
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilder_no_masters(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7),
        ])
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=7,
                name='builder-7',
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilder_with_masters(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7),
            fakedb.Master(id=3, name='m1'),
            fakedb.Master(id=4, name='m2'),
            fakedb.BuilderMaster(builderid=7, masterid=3),
            fakedb.BuilderMaster(builderid=7, masterid=4),
        ])
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(
            builderdict,
            builders.BuilderModel(
                id=7,
                name='builder-7',
                masterids=[3, 4],
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilder_missing(self):
        builderdict = yield self.db.builders.getBuilder(7)
        self.assertEqual(builderdict, None)

    @defer.inlineCallbacks
    def test_getBuilders(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7, name='some:builder'),
            fakedb.Builder(id=8, name='other:builder'),
            fakedb.Builder(id=9, name='third:builder'),
            fakedb.Master(id=3, name='m1'),
            fakedb.Master(id=4, name='m2'),
            fakedb.BuilderMaster(builderid=7, masterid=3),
            fakedb.BuilderMaster(builderid=8, masterid=3),
            fakedb.BuilderMaster(builderid=8, masterid=4),
        ])
        builderlist = yield self.db.builders.getBuilders()
        self.assertEqual(
            sorted(builderlist, key=builderKey),
            sorted(
                [
                    builders.BuilderModel(
                        id=7,
                        name='some:builder',
                        masterids=[3],
                    ),
                    builders.BuilderModel(
                        id=8,
                        name='other:builder',
                        masterids=[3, 4],
                    ),
                    builders.BuilderModel(
                        id=9,
                        name='third:builder',
                    ),
                ],
                key=builderKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilders_masterid(self):
        yield self.db.insert_test_data([
            fakedb.Builder(id=7, name='some:builder'),
            fakedb.Builder(id=8, name='other:builder'),
            fakedb.Builder(id=9, name='third:builder'),
            fakedb.Master(id=3, name='m1'),
            fakedb.Master(id=4, name='m2'),
            fakedb.BuilderMaster(builderid=7, masterid=3),
            fakedb.BuilderMaster(builderid=8, masterid=3),
            fakedb.BuilderMaster(builderid=8, masterid=4),
        ])
        builderlist = yield self.db.builders.getBuilders(masterid=3)
        self.assertEqual(
            sorted(builderlist, key=builderKey),
            sorted(
                [
                    builders.BuilderModel(
                        id=7,
                        name='some:builder',
                        masterids=[3],
                    ),
                    builders.BuilderModel(
                        id=8,
                        name='other:builder',
                        masterids=[3, 4],
                    ),
                ],
                key=builderKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilders_projectid(self):
        yield self.db.insert_test_data([
            fakedb.Project(id=201, name="p201"),
            fakedb.Project(id=202, name="p202"),
            fakedb.Builder(id=101, name="b101"),
            fakedb.Builder(id=102, name="b102", projectid=201),
            fakedb.Builder(id=103, name="b103", projectid=201),
            fakedb.Builder(id=104, name="b104", projectid=202),
            fakedb.Master(id=3, name='m1'),
            fakedb.Master(id=4, name='m2'),
            fakedb.BuilderMaster(builderid=101, masterid=3),
            fakedb.BuilderMaster(builderid=102, masterid=3),
            fakedb.BuilderMaster(builderid=103, masterid=4),
            fakedb.BuilderMaster(builderid=104, masterid=4),
        ])
        builderlist = yield self.db.builders.getBuilders(projectid=201)
        self.assertEqual(
            sorted(builderlist, key=builderKey),
            sorted(
                [
                    builders.BuilderModel(
                        id=102,
                        name="b102",
                        masterids=[3],
                        projectid=201,
                    ),
                    builders.BuilderModel(
                        id=103,
                        name="b103",
                        masterids=[4],
                        projectid=201,
                    ),
                ],
                key=builderKey,
            ),
        )

    @async_to_deferred
    async def test_getBuilders_workerid(self):
        await self.db.insert_test_data([
            fakedb.Builder(id=101, name="b101"),
            fakedb.Builder(id=102, name="b102"),
            fakedb.Builder(id=103, name="b103"),
            fakedb.Builder(id=104, name="b104"),
            fakedb.Master(id=3, name='m1'),
            fakedb.BuilderMaster(id=1, builderid=101, masterid=3),
            fakedb.BuilderMaster(id=2, builderid=102, masterid=3),
            fakedb.BuilderMaster(id=3, builderid=103, masterid=3),
            fakedb.BuilderMaster(id=4, builderid=104, masterid=3),
            fakedb.Worker(id=1, name='zero'),
            fakedb.ConnectedWorker(id=1, workerid=1, masterid=3),
            fakedb.ConfiguredWorker(id=1, workerid=1, buildermasterid=2),
            fakedb.ConfiguredWorker(id=2, workerid=1, buildermasterid=3),
        ])
        builderlist = await self.db.builders.getBuilders(workerid=1)
        self.assertEqual(
            sorted(builderlist, key=builderKey),
            sorted(
                [
                    builders.BuilderModel(
                        id=102,
                        name="b102",
                        masterids=[3],
                    ),
                    builders.BuilderModel(
                        id=103,
                        name="b103",
                        masterids=[3],
                    ),
                ],
                key=builderKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuilders_empty(self):
        builderlist = yield self.db.builders.getBuilders()
        self.assertEqual(sorted(builderlist), [])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot_worker.commands import registry
from buildbot_worker.commands import shell


class Registry(unittest.TestCase):
    def test_getFactory(self) -> None:
        factory = registry.getFactory('shell')
        self.assertEqual(factory, shell.WorkerShellCommand)

    def test_getFactory_KeyError(self) -> None:
        with self.assertRaises(KeyError):
            registry.getFactory('nosuchcommand')

    def test_getAllCommandNames(self) -> None:
        self.assertTrue('shell' in registry.getAllCommandNames())

    def test_all_commands_exist(self) -> None:
        # if this doesn't raise a KeyError, then we're good
        for n in registry.getAllCommandNames():
            registry.getFactory(n)



# This file is part of .  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright  Team Members

from __future__ import annotations

import json
from typing import TYPE_CHECKING
from typing import Any

from autobahn.twisted.resource import WebSocketResource
from autobahn.twisted.websocket import WebSocketServerFactory
from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.websocket.types import ConnectionDeny
from autobahn.websocket.types import ConnectionRequest
from twisted.internet import defer
from twisted.python import log

try:
    from twisted.internet.interfaces import ISSLTransport

    _HAS_SSL = True
except ImportError:
    _HAS_SSL = False


from buildbot.util import bytes2unicode
from buildbot.util import toJson
from buildbot.www import auth

if TYPE_CHECKING:
    from buildbot.master import BuildMaster
    from buildbot.mq.base import QueueRef
    from buildbot.util.twisted import InlineCallbacksType


class Subscription:
    def __init__(self, query: str, id: str):
        self.query = query
        self.id = id
        self.last_value_chksum = None


def parse_cookies(header_value: str) -> dict[str, str]:
    cookies: dict[str, str] = {}

    # autobahn appends multiple cookies using ','
    for kv1 in header_value.split(","):
        for kv in kv1.split(";"):
            kv = kv.lstrip()
            try:
                k, v = kv.split("=", 1)
                cookies[k] = v
            except ValueError:
                pass
    return cookies


def get_ws_sitepath(path: str) -> list[bytes]:
    parts = path.split('/')
    if len(parts) > 0 and parts[-1] == 'ws':
        parts.pop()
    return [p.encode('utf-8') for p in parts if p]


class WsProtocol(WebSocketServerProtocol):
    def __init__(self, master: BuildMaster):
        super().__init__()
        self.master = master
        self.qrefs: dict[str, QueueRef] | None = {}
        self.debug = self.master.config.www.get("debug", False)

    def to_json(self, msg: dict[str, Any]) -> bytes:
        return json.dumps(msg, default=toJson, separators=(",", ":")).encode()

    def send_json_message(self, **msg: Any) -> defer.Deferred:
        return self.sendMessage(self.to_json(msg))

    def send_error(self, error: str, code: int, _id: str | None) -> defer.Deferred:
        return self.send_json_message(error=error, code=code, _id=_id)

    def onMessage(self, frame: bytes, isBinary: bool) -> defer.Deferred | None:
        """
        Parse the incoming request.
        """
        if self.debug:
            log.msg(f"FRAME {bytes2unicode(frame)}")

        frame_dict = json.loads(bytes2unicode(frame))
        _id = frame_dict.get("_id")
        _type = frame_dict.pop("type", None)
        if _id is None and _type is None:
            return self.send_error(
                error="no '_id' or 'type' in websocket frame", code=400, _id=None
            )

        cmd = frame_dict.pop("cmd", None)
        if cmd is None:
            return self.send_error(error="no 'cmd' in websocket frame", code=400, _id=None)
        cmdmeth = "cmd_" + cmd

        meth = getattr(self, cmdmeth, None)
        if meth is None:
            return self.send_error(error=f"no such command type '{cmd}'", code=404, _id=_id)
        try:
            return meth(**frame_dict)
        except TypeError as e:
            return self.send_error(error=f"Invalid method argument '{e!s}'", code=400, _id=_id)
        except Exception as e:
            log.err(e, f"while calling command {cmdmeth}")
            return self.send_error(error=f"Internal Error '{e!s}'", code=500, _id=_id)

    # legacy protocol methods

    def ack(self, _id: str) -> defer.Deferred:
        return self.send_json_message(msg="OK", code=200, _id=_id)

    def parsePath(self, path: str) -> tuple[str | None, ...]:
        path_parts = path.split("/")
        return tuple(str(p) if p != "*" else None for p in path_parts)

    def isPath(self, path: Any) -> bool:
        if not isinstance(path, str):
            return False
        return True

    @defer.inlineCallbacks
    def cmd_startConsuming(self, path: str, _id: str) -> InlineCallbacksType[None]:
        if not self.isPath(path):
            yield self.send_json_message(error=f"invalid path format '{path!s}'", code=400, _id=_id)
            return

        # if it's already subscribed, don't leak a subscription
        if self.qrefs is not None and path in self.qrefs:
            yield self.ack(_id=_id)
            return

        def callback(key: list[str], message: Any) -> defer.Deferred:
            # protocol is deliberately concise in size
            return self.send_json_message(k="/".join(key), m=message)

        qref = yield self.master.mq.startConsuming(callback, self.parsePath(path))

        # race conditions handling
        if self.qrefs is None or path in self.qrefs:
            qref.stopConsuming()

        # only store and ack if we were not disconnected in between
        if self.qrefs is not None:
            self.qrefs[path] = qref
            self.ack(_id=_id)

    @defer.inlineCallbacks
    def cmd_stopConsuming(self, path: str, _id: str) -> InlineCallbacksType[None]:
        if not self.isPath(path):
            yield self.send_json_message(error=f"invalid path format '{path!s}'", code=400, _id=_id)
            return

        # only succeed if path has been started
        if self.qrefs is not None and path in self.qrefs:
            qref = self.qrefs.pop(path)
            yield qref.stopConsuming()
            yield self.ack(_id=_id)
            return
        yield self.send_json_message(error=f"path was not consumed '{path!s}'", code=400, _id=_id)

    def cmd_ping(self, _id: str) -> None:
        self.send_json_message(msg="pong", code=200, _id=_id)

    def connectionLost(self, reason: Any) -> None:
        if self.debug:
            log.msg("connection lost", system=self)
        if self.qrefs is not None:
            for qref in self.qrefs.values():
                qref.stopConsuming()

        self.qrefs = None  # to be sure we don't add any more

    def is_secure(self) -> bool:
        return _HAS_SSL and ISSLTransport.providedBy(self.transport)

    @defer.inlineCallbacks
    def onConnect(self, request: ConnectionRequest) -> InlineCallbacksType[None]:
        www = self.master.www
        sitepath = get_ws_sitepath(request.path)

        cookies = parse_cookies(request.headers.get('cookie', ''))
        token = cookies.get(auth.build_cookie_name(self.is_secure(), sitepath).decode('utf-8'))

        try:
            if token is None:
                user_info = auth.build_anonymous_user_info()
            else:
                user_info = auth.parse_user_info_from_token(token, www.site.session_secret)

            # assume that if user cannot access /masters endpoint, then it can't access anything
            yield www.authz.assertUserAllowed('masters', 'get', {}, user_info)
        except Exception as e:
            raise ConnectionDeny(403, "Forbidden") from e


class WsProtocolFactory(WebSocketServerFactory):
    def __init__(self, master: BuildMaster):
        super().__init__()
        self.master = master
        pingInterval = self.master.config.www.get("ws_ping_interval", 0)
        self.setProtocolOptions(webStatus=False, autoPingInterval=pingInterval)

    def buildProtocol(self, addr: Any) -> WsProtocol:
        p = WsProtocol(self.master)
        p.factory = self
        return p


class WsResource(WebSocketResource):
    def __init__(self, master: BuildMaster):
        super().__init__(WsProtocolFactory(master))
